/*
 * Copyright (c) 2005 The University of Reading
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Reading, nor the names of the
 *    authors or contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package uk.ac.rdg.resc.jstyx.messages;

import org.apache.mina.common.ByteBuffer;

import org.apache.mina.protocol.ProtocolViolationException;
import org.apache.mina.protocol.ProtocolEncoderOutput;

import uk.ac.rdg.resc.jstyx.StyxUtils;
import uk.ac.rdg.resc.jstyx.types.ULong;

/**
 * Message sent to write data to a file
 * @todo make count an int, and use MINA's ByteBuffer
 *
 * @author Jon Blower
 * $Revision: 1.9 $
 * $Date: 2005/03/22 17:48:27 $
 * $Log: TwriteMessage.java,v $
 * Revision 1.9  2005/03/22 17:48:27  jonblower
 * Removed debug code that tracked ByteBuffer allocation
 *
 * Revision 1.8  2005/03/21 17:57:11  jonblower
 * Trying to fix ByteBuffer leak in SGS server
 *
 * Revision 1.7  2005/03/17 07:29:36  jonblower
 * Fixed bug in getData()
 *
 * Revision 1.6  2005/03/16 22:16:43  jonblower
 * Added Styx Grid Service classes to core module
 *
 * Revision 1.5  2005/03/16 17:56:22  jonblower
 * Replaced use of java.nio.ByteBuffer with MINA's ByteBuffer to minimise copying of buffers
 *
 * Revision 1.4  2005/03/15 09:01:48  jonblower
 * Message type now stored as short, not int
 *
 * Revision 1.3  2005/03/11 14:02:16  jonblower
 * Merged MINA-Test_20059309 into main line of development
 *
 * Revision 1.2.2.2  2005/03/11 12:30:46  jonblower
 * Changed so that message payloads are always ints, not longs
 *
 * Revision 1.2.2.1  2005/03/10 11:50:59  jonblower
 * Changed to fit with MINA framework
 *
 * Revision 1.2  2005/02/24 07:44:44  jonblower
 * Added getFriendlyString()
 *
 * Revision 1.1.1.1  2005/02/16 18:58:30  jonblower
 * Initial import
 *
 */
public class TwriteMessage extends StyxMessage
{
    
    private short fid;     // The fid to write data to
    private ULong offset; // The offset in the file to which to write data
    private int count;   // The number of bytes to write to the file
    
    private ByteBuffer data; // Buffer containing the data to write to the file.
                             // This is only used if we have decoded this message
                             // from bytes that have been read over the network
                             // (i.e. only used by server programs)
    private int dataPos;     // Stores the location of the first byte of payload
                             // data in the ByteBuffer
    
    private byte[] bytes; // The raw data bytes (only used if we have constructed
                          // this Tmessage from scratch before sending it).  It is
                          // not used if this message is decoded from bytes that
                          // have been read over the network. I.e. this is only
                          // used by client programs.
    private int pos;      // If the byte array is filled, this is the index of
                          // the first byte in the array to write
    
    /** 
     * Creates a new TwriteMessage. This constructor will be called by the
     * MessageRecognizer.
     * @param length The total length of the message (including all header info)
     * @param type The type of the message (a number between 100 and 127)
     * @param tag The tag that identifies this message
     */
    public TwriteMessage(int length, int tag)
    {
        super(length, (short)16, tag);
        this.name = "Twrite";
        this.bytes = null;
    }
    
    /**
     * Creates a new TwriteMessage that will write the given bytes to the 
     * file described by the given fid at the given offset. N.B. this will
     * use all of the bytes in the given array.
     */
    public TwriteMessage(short fid, ULong offset, byte[] bytes)
    {
        this(fid, offset, bytes, 0, bytes.length);
    }
    
    /**
     * Creates a new TwriteMessage that will write the given bytes to the 
     * file described by the given fid at the given offset.
     */
    public TwriteMessage(short fid, ULong offset, byte[] bytes, int pos, int count)
    {
        this(0,  0); // The tag and length will be set later
        this.fid = fid;
        this.offset = offset;
        this.bytes = bytes; // We don't create a ByteBuffer for these bytes
        this.pos = pos;
        this.count = count;
        this.length = StyxUtils.HEADER_LENGTH + 2 + 8 + 2 + 1 + this.count;
    }
    
    /**
     * @throws ProtocolViolationException if the payload of the message is more
     * than Integer.MAX_VALUE
     */
    protected final void decodeBody(StyxBuffer buf)
        throws ProtocolViolationException
    {
        this.fid = buf.getUShort();
        this.offset = buf.getULong();
        long n = buf.getUShort();
	byte ignore = buf.getUByte();
        if (n < 0 || n > Integer.MAX_VALUE)
        {
            throw new ProtocolViolationException("Payload of Rread message " +
                "cannot be less than 0 or greater than Integer.MAX_VALUE bytes");
        }
        this.count = (int)n; // We know this cast must be safe
        this.length = StyxUtils.HEADER_LENGTH + 2 + 8 + 2 + 1 + this.count;
	System.out.println("write count = "+this.count);
        
	/*
	  // this is taken care of in StyxMessage.decode
        this.data = ByteBuffer.allocate(this.count);
	byte[] b = buf.getData(this.count);
	this.data.put(b);
	this.dataPos = 0;
	*/
    }
    
    
    /**
     * Writes the message into the given StyxBuffer. This method should only be
     * called by client programs, i.e. those that have created the TwriteMessage
     * "from scratch" from a byte array. Otherwise, this will throw an 
     * IllegalStateException
     */
    protected final void encodeBody(StyxBuffer buf)
    {
        buf.putUShort(this.fid).putULong(this.offset).putUShort(this.count);
	buf.putUByte(0); //pad
        if (this.bytes != null)
        {
            buf.put(this.bytes, this.pos, this.count);
        }
        else
        {
            // If this.data were non-null, we wouldn't have called this method.
            throw new IllegalStateException("This RreadMessage contains no data");
        }
    }
    
    public short getFid()
    {
        return this.fid;
    }
    
    public ULong getOffset()
    {
        return this.offset;
    }
    
    public int getCount()
    {
        return this.count;
    }
    

    public ByteBuffer getRawData()
    {
        if (this.data == null)
        {
            throw new IllegalStateException("Data buffer is null");
        }
        this.data.position(this.dataPos);
        return this.data;
    }
    
    
    /**
     * @return The data that will be written to the file (i.e. the message's
     * payload, as java.nio.ByteBuffer with the position and limit set correctly.
     * The java.nio.ByteBuffer that is returned is created simply by wrapping
     * the byte array that was used to create this message in the first place,
     * therefore the array() method of the ByteBuffer can always be used.
     *
     * This method should <b>only</b> be called by clients (i.e. a program that
     * created this TwriteMessage from a byte array), otherwise this method
     * will throw an IllegalStateException.
     */
    public org.apache.mina.common.ByteBuffer getData()
    {
        if (this.data == null)
        {
            throw new IllegalStateException("Byte array is null");
        }
        return this.data;
    }

    public void setData(org.apache.mina.common.ByteBuffer buff)
    {
        if (buff == null)
        {
            throw new IllegalStateException("Byte array is null");
        }
	this.data = buff;

    }

    public void setDataPos(int pos)    
    {
	this.dataPos = pos;
    }

    protected String getElements()
    {
        StringBuffer s = new StringBuffer(", " + this.fid + ", " + this.offset +
            ", " + this.count + ", ");
        if (this.data == null)
        {
            s.append(StyxUtils.getDataSummary(30, this.bytes));
        }
        else
        {
            s.append(StyxUtils.getDataSummary(30, this.getRawData()));
        }
        return s.toString();
    }
    
    public String toFriendlyString()
    {
        StringBuffer s = new StringBuffer("fid: ");
        s.append(this.fid);
        s.append(", offset: ");
        s.append(this.offset);
        s.append(", count: ");
        s.append(this.count);
        s.append(", ");
        if (this.data == null)
        {
            s.append(StyxUtils.getDataSummary(30, this.bytes));
        }
        else
        {
            s.append(StyxUtils.getDataSummary(30, this.getRawData()));
        }
        return s.toString();
    }
    
}
