/*
 * Copyright (c) 2005 The University of Reading
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Reading, nor the names of the
 *    authors or contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package uk.ac.rdg.resc.jstyx.server;

import net.tinyos.util.*;
import net.tinyos.message.*;
import net.tinyos.tools.*;
import java.util.*;
import java.lang.Thread;

import org.apache.log4j.Logger;

import org.apache.mina.common.ByteBuffer;

import uk.ac.rdg.resc.jstyx.StyxUtils;
import uk.ac.rdg.resc.jstyx.StyxException;
import uk.ac.rdg.resc.jstyx.types.ULong;

/**
 * File whose underlying data are stored as a ByteBuffer in memory. This buffer
 * can grow to arbitrary size.
 *
 * @author Jon Blower
 * $Revision: 1.19 $
 * $Date: 2005/09/08 07:08:59 $
 * $Log: InMemoryFile.java,v $
 * Revision 1.19  2005/09/08 07:08:59  jonblower
 * Removed "String user" from list of parameters to StyxFile.write()
 *
 * Revision 1.18  2005/09/02 16:51:20  jonblower
 * Fixed bug with ByteBuffers being released prematurely and changed to use autoExpanding ByteBuffers
 *
 * Revision 1.17  2005/09/01 07:50:22  jonblower
 * Trying to fix bug with ByteBuffers being released (can't read from InMemoryFile)
 *
 * Revision 1.16  2005/08/30 16:29:00  jonblower
 * Added processAndReplyRead() helper functions to StyxFile
 *
 * Revision 1.13  2005/06/10 07:53:12  jonblower
 * Changed SGS namespace: removed "inurl" and subsumed functionality into "stdin"
 *
 * Revision 1.12  2005/04/28 08:11:15  jonblower
 * Modified permissions handling in documentation directory of SGS
 *
 * Revision 1.10  2005/04/27 16:11:43  jonblower
 * Added capability to add documentation files to SGS namespace
 *
 * Revision 1.9  2005/04/26 07:46:11  jonblower
 * Continuing to improve setting of parameters in Styx Grid Services
 *
 * Revision 1.8  2005/03/24 15:11:07  jonblower
 * Changed so that underlying ByteBuffer is allocated on the first write to the file
 *
 * Revision 1.7  2005/03/24 13:07:09  jonblower
 * Added code to prevent the file growing to larger than a specified size
 *
 * Revision 1.6  2005/03/24 12:55:14  jonblower
 * Changed to use ByteBuffer as backing store
 *
 * Revision 1.5  2005/03/24 09:48:31  jonblower
 * Changed 'count' from long to int throughout for reading and writing
 *
 * Revision 1.4  2005/03/16 17:56:23  jonblower
 * Replaced use of java.nio.ByteBuffer with MINA's ByteBuffer to minimise copying of buffers
 *
 * Revision 1.3  2005/03/11 14:02:16  jonblower
 * Merged MINA-Test_20059309 into main line of development
 *
 * Revision 1.2.2.1  2005/03/10 20:55:38  jonblower
 * Removed references to Netty
 *
 * Revision 1.2  2005/03/01 13:47:43  jonblower
 * Changed default user and group to 'user' and 'group'
 *
 * Revision 1.1.1.1  2005/02/16 18:58:31  jonblower
 * Initial import
 *
 */
public class EventFile extends SensorFile implements EventResponse
{    

    public class BufferedResponse
    {
	StyxFileClient client;
	long offset;
	int count;
        int tag;
	short fid;

	public BufferedResponse(StyxFileClient client, long offset, int count, int tag, short fid)
	{
	    this.client = client;
	    this.offset = offset;
	    this.count = count;
	    this.tag = tag;
	    this.fid = fid;
	}
    }

    public static final byte ASCIIZERO = 48;

    /* all the queued up messages that we need to respond to once we hear an event */
    private Vector ResponseVec;

    /* value corresponding to the last event, -1 if invalid    */
    byte lastvalue;

    /**
     * Creates a new instance of EventFile.
     */

    public EventFile(String name, int sensorid, EPMote epm, int rid)
        throws StyxException
    {
        super(name, sensorid, epm, rid);
	ResponseVec = new Vector();
	lastvalue = -1;
	this.getMote(sensorid).registerEventHandler(this);
    }



	
    /* need to handle error better here  */
    public synchronized void read(StyxFileClient client, long offset, int count,
        int tag, short fid) throws StyxException
    {
	/* create a new packet to send off ... */
	byte buffer[];
	int size=10;
	buffer=new byte[size];
	int retvalue;

	EPMsg outpacket = new EPMsg(buffer);
	EPMsgR returnpacket = new EPMsgR(); 
	
	retvalue = 0;
	if (offset > 0) //second read in sequence, reply with nothing more
	    this.processAndReplyRead(java.lang.String.valueOf(retvalue), client, offset, 0, tag, fid);	
	if (lastvalue != -1) {
	    
	    outpacket.set_arg((byte)0);  // irrelevant
	    outpacket.set_type(SNFS.S_Tread); //define operation to be 'read'
	    outpacket.set_rid((byte)(this.rid));  	
	    try {
		this.getMote(this.sensorid).sendrecv(outpacket, returnpacket);
	    }
	    catch(Exception e) 
		{
		    e.printStackTrace();
		}


	    
	    retvalue = lastvalue;
	    System.out.println("immediate return of read event\n");
	    this.processAndReplyRead(java.lang.String.valueOf(retvalue), client, offset, count, tag, fid);	
	    lastvalue = -1;
	}
	 else
	    {
		BufferedResponse br = new BufferedResponse(client, offset, count, tag, fid);
		System.out.println("no read event, stalling request\n");
		ResponseVec.add(br);
	    }
    } 
		
    
    public synchronized void write(StyxFileClient client, long offset,
        int count, ByteBuffer data, boolean truncate, int tag)
        throws StyxException
    {
	/* create a new packet to send off ... */
	byte buffer[];
	int size=10;
	byte databuffer[];

	int retvalue;
	buffer=new byte[size];
	databuffer=new byte[data.remaining()];
	EPMsg outpacket = new EPMsg(buffer);
	EPMsgR returnpacket = new EPMsgR(); 
	
	
	    // if (offset == 0) this is a trace ON/OFF request, otherwise set threshold ...

	if ((offset != 0) && (offset != 1)) {
	    System.out.println("EventFile write: arg is not 0/1\n");
	    return;
	}
	outpacket.set_offset((byte)offset);
	    
	// the argument is either '1' or '0', determining trace ON/OFF or 1 byte threshold
	data.get(databuffer);
	//	outpacket.set_arg((byte)data.toByteArray()[0]); 
	outpacket.set_arg((byte)(databuffer[0] - ASCIIZERO)); 
	
	outpacket.set_type(SNFS.S_Twrite); //define operation to be 'write'
	outpacket.set_rid((byte)(this.rid));  	
	System.out.println("evwrite : rid ="+ this.rid+ ", offset = "+offset+", val = "+(byte)(databuffer[0] - ASCIIZERO) + " databuf[0] = "+databuffer[0]);

	try {
	    this.getMote(this.sensorid).sendrecv(outpacket, returnpacket);
	    /*   if an error is returned need to do something  */
	}
	catch(Exception e) 
	    {
		e.printStackTrace();
	    }

	if (returnpacket.get_type() == SNFS.S_Rwrite)
	    System.out.println("Eventfile write complete\n");

	if (returnpacket.get_type() == SNFS.S_Rerror)
	    System.out.println("error conventional writing to Eventfile\n");

	if (returnpacket.get_type() == SNFS.S_Rnoid)
	    System.out.println("error - no such resource writing to Eventfile\n");

        this.replyWrite(client, count, tag);

    }

    public int getsensorid() {
	return super.getsensorid();
    }


    public int getrid() {
	return super.getrid();
    }

    public void Response(byte retvalue)
    {
	if (!ResponseVec.isEmpty()) { // we have requests waiting for this value
	    Enumeration e = ResponseVec.elements();
	    while (e.hasMoreElements()) {
		BufferedResponse br = (BufferedResponse) e.nextElement();
		this.processAndReplyRead(java.lang.String.valueOf(retvalue), br.client, br.offset, br.count, br.tag, br.fid);		    
		System.out.println("event came in, releasing request\n");
	    }
	    ResponseVec.clear();
	}
	else {
	    System.out.println("event with no takers, buffering\n");
	    lastvalue = retvalue; // no requests waiting, just buffer it up 
	}
    }
    
 
    
    
}
