/*
 * Copyright (c) 2005 The University of Reading
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Reading, nor the names of the
 *    authors or contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package uk.ac.rdg.resc.jstyx.server;

import net.tinyos.util.*;
import net.tinyos.message.*;
import net.tinyos.tools.*;
import net.tinyos.deluge.*;
import java.util.*;
import java.lang.Thread;

import org.apache.log4j.Logger;

import org.apache.mina.common.ByteBuffer;

import uk.ac.rdg.resc.jstyx.StyxUtils;
import uk.ac.rdg.resc.jstyx.StyxException;
import uk.ac.rdg.resc.jstyx.types.ULong;

/**
 * File whose underlying data are stored as a ByteBuffer in memory. This buffer
 * can grow to arbitrary size.
 *
 * @author Jon Blower
 * $Revision: 1.19 $
 * $Date: 2005/09/08 07:08:59 $
 * $Log: InMemoryFile.java,v $
 * Revision 1.19  2005/09/08 07:08:59  jonblower
 * Removed "String user" from list of parameters to StyxFile.write()
 *
 * Revision 1.18  2005/09/02 16:51:20  jonblower
 * Fixed bug with ByteBuffers being released prematurely and changed to use autoExpanding ByteBuffers
 *
 * Revision 1.17  2005/09/01 07:50:22  jonblower
 * Trying to fix bug with ByteBuffers being released (can't read from InMemoryFile)
 *
 * Revision 1.16  2005/08/30 16:29:00  jonblower
 * Added processAndReplyRead() helper functions to StyxFile
 *
 * Revision 1.13  2005/06/10 07:53:12  jonblower
 * Changed SGS namespace: removed "inurl" and subsumed functionality into "stdin"
 *
 * Revision 1.12  2005/04/28 08:11:15  jonblower
 * Modified permissions handling in documentation directory of SGS
 *
 * Revision 1.10  2005/04/27 16:11:43  jonblower
 * Added capability to add documentation files to SGS namespace
 *
 * Revision 1.9  2005/04/26 07:46:11  jonblower
 * Continuing to improve setting of parameters in Styx Grid Services
 *
 * Revision 1.8  2005/03/24 15:11:07  jonblower
 * Changed so that underlying ByteBuffer is allocated on the first write to the file
 *
 * Revision 1.7  2005/03/24 13:07:09  jonblower
 * Added code to prevent the file growing to larger than a specified size
 *
 * Revision 1.6  2005/03/24 12:55:14  jonblower
 * Changed to use ByteBuffer as backing store
 *
 * Revision 1.5  2005/03/24 09:48:31  jonblower
 * Changed 'count' from long to int throughout for reading and writing
 *
 * Revision 1.4  2005/03/16 17:56:23  jonblower
 * Replaced use of java.nio.ByteBuffer with MINA's ByteBuffer to minimise copying of buffers
 *
 * Revision 1.3  2005/03/11 14:02:16  jonblower
 * Merged MINA-Test_20059309 into main line of development
 *
 * Revision 1.2.2.1  2005/03/10 20:55:38  jonblower
 * Removed references to Netty
 *
 * Revision 1.2  2005/03/01 13:47:43  jonblower
 * Changed default user and group to 'user' and 'group'
 *
 * Revision 1.1.1.1  2005/02/16 18:58:31  jonblower
 * Initial import
 *
 */
public class RadioCtlFile extends SensorFile
{    

    public static final byte ASCIIZERO = 48;

    public static final String xmlfilepath = "/cygwin/opt/tinyos-1.x/images/imagefile.xml";

    //    public static final String tmotebsl = "tmote-bsl -e -p C:/cygwin/opt/moteiv/tos/lib/Deluge/TOSBoot/build/tmote/main.ihex -p /cygwin/opt/tinyos-1.x/images/imagefile -r --telosb";
    public static final String delugeprgm = "java net.tinyos.tools.Deluge --inject --tosimage="+xmlfilepath+" --imgnum=5";
    public static final String delugereset = "java net.tinyos.tools.Deluge --reboot --imgnum=6";

    private int network_address;

    EPMote mote;

    /**
     * Creates a new instance of RadioCtlFile.
     */

    public RadioCtlFile(String name, int sensorid, EPMote epm, int rid)
        throws StyxException
    {
        super(name, sensorid, epm, rid);
	this.network_address = sensorid;
    }
	
    /* need to handle error better here  */
    public synchronized void read(StyxFileClient client, long offset, int count,
        int tag, short fid) throws StyxException
    {

	int retvalue = 0;
	this.processAndReplyRead(java.lang.String.valueOf(retvalue), client, offset, 0, tag, fid);	

    } 

    public EPMote getMote(int index) {
	return this.mote;
    } 

    public synchronized void write(StyxFileClient client, long offset,
        int count, ByteBuffer data, boolean truncate, int tag)
        throws StyxException
    {
	byte databuffer[];
	Runtime rt;
	Process dwld;

	databuffer=new byte[data.remaining()];
	data.get(databuffer);

	String writedata = new String(databuffer, 0, count);
	writedata.trim();
	System.out.println("write count = "+count);

	if (writedata.startsWith("mac")) {
	    System.out.println("mac cmd = "+writedata);
	    writedata = writedata.substring(4);
	    writedata.trim();	
	    System.out.println("reassgning wireless address : " + writedata);
	    this.getMote(this.sensorid).set_moteid(Integer.valueOf(writedata).intValue());
	    this.replyWrite(client, count, tag);
	    return;
	}

	/* program command */
	if (databuffer[0] == (byte) 112) {
	
	System.out.println("RadioCtlFile write\n");
        this.replyWrite(client, count, tag);

	try {
	    rt = Runtime.getRuntime();
	}
	catch(Exception e) 
	    {
		System.out.println("unable to get runtime");
		return;
	    }		

	try {
	    //	    dwld = rt.exec(tmotebsl+" -c "+this.comport);
	    System.out.println("deluge prgming cmd: "+ delugeprgm);
	    dwld = rt.exec(delugeprgm);
	    System.out.println("deluge reset cmd: "+ delugereset);
	    dwld = rt.exec(delugereset);

	}
	catch(Exception e) 
	    {
		System.out.println("unable to exec deluge");
		return;
	    }
	try {
	    dwld.waitFor();
	}
	catch(Exception e) 
	    {
		System.out.println("unable to wait for completion");
		return;
	    }
        this.replyWrite(client, count, tag);
	return;
	}


	if (writedata.startsWith("execute")) {

	    writedata = writedata.substring("execute".length() + 1); /* 12 chars for PROGRAMIMAGE + ' ' */
	    writedata.trim();	
	    System.out.println("programming sensor <"+this.network_address + "> with image num:"+ writedata);

	    try {
		 DelugeAdvMsg advMsg = new DelugeAdvMsg();
		 
		 advMsg.set_sourceAddr(0);
		 advMsg.set_version(DelugeConsts.DELUGE_VERSION);
		 advMsg.set_type((short)DelugeConsts.DELUGE_SINGLE_RESET);
		 advMsg.set_nodeDesc_vNum((short)0);
		 advMsg.set_nodeDesc_uid(0);
		 advMsg.set_nodeDesc_imgNum((short)this.network_address);
		 advMsg.set_imgDesc_vNum((short)0);
		 advMsg.set_imgDesc_imgNum((short)Integer.valueOf(writedata).intValue()); /* image number to reprogram with */
		 this.getMote(this.sensorid).send(advMsg);

	    }
	
	    catch(Exception e) 
		{
		    System.out.println("unable to send deluge reboot packet");
		    System.out.println("exception : "+e);
		    return;
		}

	    this.replyWrite(client, count, tag);
	    return;
	}
    }


	


    public int getsensorid() {
	return super.getsensorid();
    }

    public int getrid() {
	return super.getrid();
    }

    
 
    
    
}
