/*========================================================================*\
|                                                                          |
| Flash.c                                                                  |
|                                                                          |
| This file contains the FLASH memory functions (set device speed, erase,  |
| and program).                                                            |
| The FLASH is erased/programmed by a stub program (and data) which is     |
| downloaded into the device RAM, and which is then executed. The stubs    |
| manipulate the flash controller in "user space".                         |
|--------------------------------------------------------------------------|
| Project:              MSP430 JTAG interfcae                              |
| Developed using:      MS Visual C++ 5.0                                  |
|--------------------------------------------------------------------------|
| Author:               Dale Wellborn                                      |
| Version:              1.03                                               |
| Initial Version:      20 / 07 / 02                                       |
| Last Change:                                                             |
|--------------------------------------------------------------------------|
| Version history:                                                         |
| Version: 1.0.1                                                           |
| 15.11.2002 UPSF restore of DCO registers added                           |
| Version: 1.0.2                                                           |
| 04.02.2003 UPSF restore of RAM after Flash Programming fixed             |
| Version: 1.1.3                                                           |
| 08.05.2003 UPSF fixed mass erase time to 200ms                           |
|                 changed set DCO to try it 10 times before returing error |
| 14.05.2003 UPSF changed timeout counter value for executeCode by *100    |
|                  to ensure proper working with faster PC's               |
|                 removed WDTHOLD from functions which are loaded to the   |
|                  MSP430 to ensure that the WDTNMI is not executed        |
|                 added save and restore from IE1 and IE2 to aviod an NMI  |
|                  during the Programming                                  |
| 26.08.2003 UPSF changed timeout counter value for executeCode now uses   |
|                  HIL_DelayMSec                                           |
| Version: 1.1.5                                                           |
| 01.12.2003 UPSF added writeFlash and eraseFlash function from MSP430.c   |
|                 changed some parameter to long                           |
| 21.01.2004 UPSF support for F1611                                        |
| 29.02.2004 UPSF Added setupMarginalReadFlash                             |
|                                                                          |
|                                                                          |
|--------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments                                       |
\*========================================================================*/

/*------------------------------------------------------------------------*\
| Remarks:                                                                 |
|                                                                          |
\*------------------------------------------------------------------------*/

// #includes. -----------------------------------------------------------------

#include "MSP/Basic_Types.h"
#include "MSP/HIL.h"
#include <stdlib.h>
#include <string.h>

#include "MSP/MSP430_Debug.h"

#include "MSP/Execution.h"
#include "MSP/Flash.h"
#include "MSP/Globals.h"
#include "MSP/JTAG.h"
#include "MSP/JTAG_defs.h"
#include "MSP/Memory.h"

// (File) Global variables. ---------------------------------------------------

#define MAXRAMSIZE (0x9FF - 0x200 +1)
static CHAR ram[MAXRAMSIZE];
static CHAR verifyRam[MAXRAMSIZE];
BOOL programFrequencySet = FALSE;
BOOL eraseFrequencySet = FALSE;

// (Local) Function prototypes. -----------------------------------------------

static STATUS_T setDCOPlus(void);
static LONG getDCOPlusFreq(WORD SCFI0, WORD SCFI1, WORD SCFQCTL, WORD FLL_CTL0, WORD FLL_CTL1);
static STATUS_T setDCO(void);
LONG getDCOFreq(WORD DCO, WORD BCS1);
static STATUS_T executeCode(WORD* code, WORD sizeCode, BOOL verify, BOOL wait);
static STATUS_T programFlash(LONG address, BYTE* buffer, LONG count);
static STATUS_T _eraseFlash(LONG type, LONG address, LONG length);

static STATUS_T setDeviceFrequency(void);
static STATUS_T preserveRam(void);
static STATUS_T restoreRam(void);

// Functions. -----------------------------------------------------------------

STATUS_T setupMarginalReadFlash(LONG type)
{
// -----------------------------------------------
//  Setup Marginal RESET0
// -----------------------------------------------
// RESET0
   static WORD code_RESET0[] = {
       0x40B2,0xA580,0x012C, //         mov       #FLWKEY+FLTREGEN,&FLCTL3          ;enable test register
       0x40B2,0xA50C,0x012E, //         mov       #FLWKEY+FLFA+FLTP,&FLCTL4         ;flash memory access & test pulse
       0x40B2,0xA500,0x012C, //         mov       #FLWKEY,&FLCTL3                   ;disable/reset test register!
//   };
//// RESET1
//   static WORD code_RESET1[] = {
       0x40B2,0xA580,0x012C, //         mov       #FLWKEY+FLTREGEN,&FLCTL3          ;enable test register
       0x40B2,0xA50C,0x012E, //         mov       #FLWKEY+FLFA+FLTP,&FLCTL4         ;flash memory access & test pulse
       0x40B2,0xA504,0x012E, //         mov       #FLWKEY+FLFA,&FLCTL4              ;flash memory access
       0x40B2,0xA516,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTMRST,&FLCTL4          ;test mode & reset
       0x40B2,0xA51E,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTMRST+FLTP,&FLCTL4     ;test mode & reset & test pulse
       0x40B2,0xA504,0x012E, //         mov       #FLWKEY+FLFA,&FLCTL4               ;flash memory access
       0x40B2,0xA500,0x012C, //         mov       #FLWKEY,&FLCTL3                    ;disable/reset test register!
   };
// -----------------------------------------------
//  Setup Marginal Read 0
// -----------------------------------------------
// MGR0
   static WORD code_MGR0[] = {
       0x40B2,0xA580,0x012C, //         mov       #FLWKEY+FLTREGEN,&FLCTL3           ;enable test register
       0x40B2,0xA50C,0x012E, //         mov       #FLWKEY+FLFA+FLTP,&FLCTL4          ;flash memory access & test pulse
       0x40B2,0xA504,0x012E, //         mov       #FLWKEY+FLFA,&FLCTL4               ;flash memory access
       0x40B2,0xA516,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTMRST,&FLCTL4          ;test mode & reset
       0x40B2,0xA506,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM,&FLCTL4       ;test mode
       0x40B2,0xA50E,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTP,&FLCTL4             ;test mode & test pulse
       0x40B2,0xA516,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTMRST,&FLCTL4          ;test mode & reset
       0x40B2,0xA504,0x012E, //         mov       #FLWKEY+FLFA,&FLCTL4               ;flash memory access
       0x40B2,0xA500,0x012C, //         mov       #FLWKEY,&FLCTL3                    ;disable/reset test register!
   };
// -----------------------------------------------
//  Setup Marginal Read 1
// -----------------------------------------------
// MGR1
   static WORD code_MGR1[] = {
       0x40B2,0xA580,0x012C, //         mov       #FLWKEY+FLTREGEN,&FLCTL3          ;enable test register
       0x40B2,0xA50C,0x012E, //         mov       #FLWKEY+FLFA+FLTP,&FLCTL4         ;flash memory access & test pulse
       0x40B2,0xA504,0x012E, //         mov       #FLWKEY+FLFA,&FLCTL4              ;flash memory access
       0x40B2,0xA516,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTMRST,&FLCTL4          ;set test mode & reset
       0x40B2,0xA507,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLMGR01,&FLCTL4          ;set test mode & marginal read 1 
       0x40B2,0xA50F,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTP+FLMGR01,&FLCTL4     ;set test mode & test pulse & marginal read1
       0x40B2,0xA516,0x012E, //         mov       #FLWKEY+FLFA+FLSETTM+FLTMRST,&FLCTL4          ;set test mode & reset
       0x40B2,0xA504,0x012E, //         mov       #FLWKEY+FLFA,&FLCTL4                       ;flash memory access 
       0x40B2,0xA500,0x012C, //         mov       #FLWKEY,&FLCTL3                               ;disable/reset test register!                      
   };

    int i;

    if (! deviceHasMarginalReadFlash()) RET_ERR(PARAMETER_ERR);
    MSP430_Reset(PUC_RESET, FALSE, FALSE);
    if (cycleStart() != STATUS_OK) return (STATUS_ERROR);
    switch (type)
    {
    case FLASH_MARGINAL_READ_OFF:
        for (i=0; i <= sizeof(code_RESET0)/sizeof(WORD); i ++)
    	    cycle(code_RESET0[i]); 
        break;
    case FLASH_MARGINAL_READ_0:
        for (i=0; i <= sizeof(code_MGR0)/sizeof(WORD); i ++)
    	    cycle(code_MGR0[i]); 
        break;
    case FLASH_MARGINAL_READ_1:
        for (i=0; i <= sizeof(code_MGR1)/sizeof(WORD); i ++)
    	    cycle(code_MGR1[i]); 
        break;
    default: 
        RET_ERR(PARAMETER_ERR);
        //break;
    }
    return(cycleEnd());
};



STATUS_T setDeviceFrequency(void)
{

	STATUS_T Status;
	int i=0;
	do
	{
		i++;
		if (deviceHasDCOPlus())
		{
			Status = setDCOPlus(); // DCO+.
		}
		else
		{
			Status = setDCO(); // DCO.
		}
    }while ((i < 10) && (Status != STATUS_OK)); // try 10 times before returning error
	return (Status);
}

STATUS_T _eraseFlash(LONG type, LONG address, LONG length)
{
	// Erase Flash using cycles (and ignoring BUSY). BUSY does not work on early devices.
	static WORD code[] = {
		// BUG: The (60k) and 4k requires 20 times the number of cycles when performing a Mass Erase.
		//      For code simplicity, the number of cycles was increased by 20 for all devices.
		                      // EraseCycles     .set    5296 * 5        ;Number of clock cycles required to erase the FLASH.
		0x0200,               //                 .word   $               ;Load address.
		0x020e,               //                 .word   Main            ;Starting address.
		0x0266,               //                 .word   End             ;Ending address.
		0x0000,               // valSR           .word   0               ;Dummy value for status register.
		0xa500,               // valFCTL1        .word   0a500h          ;Password for FCTL1 register.
		0x0000,               // valADDR         .word   0               ;Dummy value for (page) address to erase.
		0x0000,               // valLEN          .word   0               ;Dummy value for length to erase (words).
//		0x40b2,0x5a80,0x0120, // Main            mov     #WDTPW+WDTHOLD,&WDTCTL ;Disable Watchdog.
		0xc232,0xc232,0xc232, //                 ; dummy stop WDT        ;Dummy instructions to replace HoldWDT (should already be stopped)
		0xc232,               //                 dint                    ;Disable interrupts.
		0x4012,0xffee,        //                 mov     valSR,SR        ;Initialize status register (Possibly disable FLL+).
		0x40b2,0xa544,0x012a, //                 mov     #0a544h,&FCTL2  ;Source = MCLK, DIV = 5.
		0x40b2,0xa500,0x012c, //                 mov     #0a500h,&FCTL3  ;Clear bits.
		0x4039,0x0014,        //                 mov     #20,R9          ;Bug requires 20 erase cycles.
		0x4092,0xffdc,0x0128, // Again           mov     valFCTL1,&FCTL1 ;Load FCTL1 value. MUST CONTAIN PASSWORD.
		0x401a,0xffd8,        //                 mov     valADDR,R10     ;Load (page) address to erase.
		0x43ba,0x0000,        //                 mov     #0ffffh,0(R10)  ;Write data into the address (to erase the FLASH).
		0x403a,0x227a,        //                 mov     #EraseCycles/3,R10 ;Delay for FLASH to erase (BUSY does not work on older devices).
		0x831a,               // Wait            dec     R10             ;Loop requires 3 clock cycles.
		0x23fe,               //                 jnz     Wait
		0x40b2,0xa500,0x0128, //                 mov     #0A500h,&FCTL1  ;Clear ERASE + MERAS bits.
		0x40b2,0xa500,0x012c, //                 mov     #0A500h,&FCTL3  ;Clear BUSY + LOCK bits.
		0x8319,               //                 dec     R9              ;Loop on erase cycles.
		0x23ed,               //                 jnz     Again 
		0x401a,0xffb8,        //                 mov     valADDR,R10     ;Load starting address.
		0x4019,0xffb6,        //                 mov     valLEN,R9       ;Load length (words).
		0x93ba,0x0000,        // Compare         cmp     #0FFFFh,0(R10)  ;Erased?
		0x2004,               //                 jne     End             ;No, exit with address of failure in R10.
		0x532a,               //                 incd    R10             ;Increment address.
		0x8319,               //                 dec     R9              ;Compare valLEN words.
		0x23fa,               //                 jnz     Compare
		0x430a,               //                 clr     R10             ;R10 contains 0 if erasure is successful.
		0x3fff,               // End             jmp     $               ;Wait here until detected by driver.
	};
	static WORD sizeCode = (WORD) (sizeof(code)/sizeof(*code));
	WORD Mass, Erase;
    long status;


	switch (type) {
	case ERASE_SEGMENT: Mass = FALSE; Erase = TRUE; break;
	case ERASE_MAIN: Mass = TRUE; Erase = FALSE; break;
	case ERASE_ALL: Mass = TRUE; Erase = TRUE; break;
	default: return (STATUS_ERROR);
	}

	// Erase the Flash.
	code[3] = (WORD) (deviceHasDCOPlus() * 0x40); // Disable the FLL+ if present.
	code[4] = (WORD) (FLASHPW << 8) | (Mass << 2) | (Erase << 1); // Combine with flash password.
	code[5] = (WORD) address;
	if (length % 2) { length++; } // Length must be an even number of bytes.
	code[6] = (WORD) (length / 2);  // Convert byte length to word length.

	if (executeCode(code, sizeCode, executeVerify, TRUE) != STATUS_OK) // Execute the Flash erase code.
	{
		return (STATUS_ERROR);
	}
	
	// R10 contains 0 if the erasure was successful, else the address of the first failing cell.
	HIL_JTAG_IR_(IR_CAPTURE_CPU_REG);
	HIL_JTAG_DR16(10); // Prepare to read R10.
    status = HIL_JTAG_DR16(10);
	return (status ? STATUS_ERROR : STATUS_OK );
}

STATUS_T programFlash(LONG address, BYTE* buffer, LONG count)
{
	static WORD code[] = {
		0x0200,               //                 .word   $               ;Load address.
		0x020c,               //                 .word   Main            ;Starting address.
		0x0260,               //                 .word   End             ;Ending address.
		0x0000,               // valSR           .word   0               ;Dummy value for status register.
		0x0000,               // valADDR         .word   0               ;Dummy value for address to program (must be even).
		0x0000,               // valLEN          .word   0               ;Dummy value for length to program (# of words).
//		0x40b2,0x5a80,0x0120, // Main            mov     #WDTPW+WDTHOLD,&WDTCTL ;Disable Watchdog.
		0xc232,0xc232,0xc232, //                 ; dummy stop WDT        ;Dummy instructions to replace HoldWDT (should already be stopped)
		0xc232,               //                 dint                    ;Disable interrupts.
		0x4012,0xfff0,        //                 mov     valSR,SR        ;Initialize status register (Possibly disable FLL+).
		0x40b2,0xa544,0x012a, //                 mov     #0a544h,&FCTL2  ;Source = MCLK, DIV = 5.
		0x40b2,0xa500,0x012c, //                 mov     #0a500h,&FCTL3  ;Clear bits.
		0x4019,0xffe2,        //                 mov     valADDR,R9      ;Address to program.
		0x4018,0xffe0,        //                 mov     valLEN,R8       ;Length to program,
		0x403a,0x0262,        //                 mov     #Buffer,R10     ;Buffer to program.
		0x9aa9,0x0000,        // Prog            cmp     @R10,0(R9)      ;Value already present?
		0x2411,               //                 jeq     Next            ;Yes, so skip and advance to next.
		0x40b2,0xa540,0x0128, //                 mov     #(0a500h | WRT),&FCTL1 ;No, so enable writing to flash.
		0x4aa9,0x0000,        //                 mov     @R10,0(R9)      ;Write data to flash.
		0xb392,0x012c,        // Wait            bit     #BUSY,&FCTL3    ;Wait for completion of write.    
		0x23fd,               //                 jnz     Wait            ;Busy bit is cleared upon completion.
		0x40b2,0xa500,0x0128, //                 mov     #0a500h,&FCTL1  ;Reset writing to flash.
		0x40b2,0xa500,0x012c, //                 mov     #0a500h,&FCTL3  ;Clear BUSY + LOCK bits.
		0x9aa9,0x0000,        //                 cmp     @R10,0(R9)      ;Verify written data.
		0x2004,               //                 jne     End             ;Verification failure (R8, R9, R10 contain details).
		0x532a,               // Next            incd    R10             ;Advance source pointer.
		0x5329,               //                 incd    R9              ;Advance destination pointer.
		0x8318,               //                 dec     R8              ;Decrement word counter.
		0x23e8,               //                 jnz     Prog            ;Program Length words.
		0x3fff                // End             jmp     $               ;Wait here until detected by driver.
							  // Buffer          .set    $               ;Start of data to program.
	};
	static WORD sizeCode = (WORD) (sizeof(code)/sizeof(*code));
	WORD deviceWords;
	WORD sizeBuffer;
	WORD *codeBuffer;
	int i, j;
	STATUS_T result;

	count /= 2; // Count words.
	deviceWords = (deviceRamEnd() + 1 - deviceRamStart()) / 2;
	sizeBuffer = (sizeCode + count) < deviceWords ? sizeCode + count : deviceWords;
	codeBuffer = (WORD*) malloc(sizeBuffer * sizeof(*codeBuffer));

	for (i = 0; i < sizeCode; i++) // Copy code into codeBuffer.
	{
		codeBuffer[i] = code[i];
	}

	codeBuffer[3] = (WORD) deviceHasDCOPlus() * 0x40; // Disable the FLL+ if present.
	codeBuffer[4] = (WORD) address;

	result = STATUS_OK;
	for (i = 0, j = 0; i < count; ) // For each data to be programmed.
	{
		codeBuffer[sizeCode + j++] = ((WORD *) buffer)[i++]; // Append data to code buffer.
		if ((i == count) || (j == sizeBuffer - sizeCode)) // End of data or code buffer full?
		{
			codeBuffer[5] = j; // Length of data to write.
			// Download (with optional Verify) and execute the program/data.
			if ((result = executeCode(codeBuffer, (WORD) (sizeCode + j), executeVerify, TRUE)) != STATUS_OK)
			{
				break; // Abort on error.
			}
			// Check result of verification. R8 contains 0 if written data verifies.
			HIL_JTAG_IR_(IR_CAPTURE_CPU_REG);
			HIL_JTAG_DR16(R8); // Prepare to read R8.
			if (HIL_JTAG_DR16(R8))
			{
				result = STATUS_ERROR;
				break; // Abort on verification error.
			}
			codeBuffer[4] = address + i * 2; // Preload next starting address.
			j = 0;
		}
	}		

	free(codeBuffer);

	return (result);
}
STATUS_T writeFlash(LONG address, BYTE* buffer, LONG count)
{ 
          // Write in flash.
              BYTE* prgBuffer = buffer;
              BOOL mallocedBuffer = FALSE;
              STATUS_T result;

              // Preserve the ram used by setDeviceFrequency() and executeCode() within programFlash().
              if (preserveRam() != STATUS_OK)
              {
                  RET_ERR(PRESERVE_RESTORE_ERR);
              }
              
              // Possibly set the device frequency.
              if (!programFrequencySet)
              {
                  if (setDeviceFrequency() != STATUS_OK)
                  {
                      RET_ERR(FREQUENCY_ERR);
                  }
                  programFrequencySet = TRUE;
              }
              
              // The flash only supports writing words on word boundaries. So force alignment if odd address
              // and/or odd count.
              // Even address and odd length: pad last, count + 1.
              if (!(address & 1) && (count & 1))
              {
                  prgBuffer = malloc(count + 1); mallocedBuffer = TRUE;
                  memcpy(prgBuffer, buffer, count);
                  if (MSP430_Read_Memory(address + count, &prgBuffer[count], 1) != STATUS_OK)
                  {
                      RET_ERR(PRESERVE_RESTORE_ERR);
                  }
                  count++;
              }
              // Odd address and even length: pad first and last, count + 2, address--.
              else if ((address & 1) && !(count & 1))
              {
                  prgBuffer = malloc(count + 2); mallocedBuffer = TRUE;
                  memcpy(prgBuffer + 1, buffer, count);
                  if (MSP430_Read_Memory(--address, &prgBuffer[0], 1) != STATUS_OK)
                  {
                      RET_ERR(PRESERVE_RESTORE_ERR);
                  }
                  if (MSP430_Read_Memory(address + count + 1, &prgBuffer[count + 1], 1) != STATUS_OK)
                  {
                      RET_ERR(PRESERVE_RESTORE_ERR);
                  }
                  count += 2;
              }
              // Odd address and odd length: pad first, count + 1, address--.
              else if ((address & 1) && (count & 1))
              {
                  prgBuffer = malloc(count + 1); mallocedBuffer = TRUE;
                  memcpy(prgBuffer + 1, buffer, count++);
                  if (MSP430_Read_Memory(--address, &prgBuffer[0], 1) != STATUS_OK)
                  {
                      RET_ERR(PRESERVE_RESTORE_ERR);
                  }
                  count++;
              }

              #ifdef DEBUG_TRACE
                    HIL_Initialize("traceoff");
              #endif
              // Program the flash (assuming that it is erased).
              result = programFlash((WORD) address, prgBuffer, (WORD) count);
              #ifdef DEBUG_TRACE
                    HIL_Initialize("traceon");
              #endif

              if (mallocedBuffer)
              {
                  free (prgBuffer);
                  mallocedBuffer = FALSE;
              }

              // Restore the ram used by setDeviceFrequency() and executeCode() within programFlash().
              if (restoreRam() != STATUS_OK)
              {
                  RET_ERR(PRESERVE_RESTORE_ERR);
              }

              if (result != STATUS_OK)
              {
                  RET_ERR(WRITE_MEMORY_ERR);
              }
     RET_OK;              
} // Write in flash.


         
STATUS_T eraseFlash(LONG type, LONG address, LONG length)
{
    STATUS_T result = STATUS_OK;

    // Preserve the ram used by setDeviceFrequency() and executeCode() within eraseFlash().
    if (preserveRam() != STATUS_OK)
    {
        RET_ERR(PRESERVE_RESTORE_ERR);
    }

    #ifdef DEBUG_TRACE
         HIL_Initialize("traceoff");
    #endif

    // Possibly set the device frequency.
    if (!eraseFrequencySet)
    {
        if (setDeviceFrequency() != STATUS_OK)
        {
            RET_ERR(FREQUENCY_ERR);
        }
        eraseFrequencySet = TRUE;
    }

    switch (type) {
    case ERASE_SEGMENT:
        {
            int segmentAddr;
            int segmentSize;

            // The following logic is "innovative" ;-) Rather than compute the segment boundaries
            // and lengths to verify, the logic steps through the segments and erases them, but it
            // only verifies one word. Once all segments have been erased, the first segment is
            // erased again and the entire length is check for erasure.

            // Erase each of the segments.
            for (segmentAddr = address; segmentAddr < address + length; segmentAddr += segmentSize)
            {
                if ((result = _eraseFlash(ERASE_SEGMENT, (WORD) segmentAddr, 1)) != STATUS_OK)
                {
                    break;
                }
            
                // The segment size is dependent upon the adress.
                if (segmentAddr < 0x1100)
                {
                    segmentSize = INFO_SEGMENT_SIZE; // Information memory.
                    continue;
                }
                if (segmentAddr < 0x1200)
                {
                    segmentSize = FIRST_60K_SEGMENT_SIZE; // The first segment of 60K devices is short.
                    continue;
                }
                segmentSize = MAIN_SEGMENT_SIZE; // Main memory.
            }
            if (result != STATUS_ERROR)
            {
                // Re-erase the first segment and check the entire length for erasure.
                result = _eraseFlash(ERASE_SEGMENT, (WORD) address, (WORD) length);
            }
        }
        break; // ERASE_SEGMENT.
    case ERASE_MAIN:
    case ERASE_ALL:
        result = _eraseFlash((WORD) type, (WORD) address, (WORD) length);
        break;
    default:
        result = STATUS_ERROR;
        break;
    }

    #ifdef DEBUG_TRACE
         HIL_Initialize("traceon");
    #endif

    // Restore the ram used by setDeviceFrequency() and executeCode() within _eraseFlash().
    if (restoreRam() != STATUS_OK)
    {
        RET_ERR(PRESERVE_RESTORE_ERR);
    }

    if (result != STATUS_OK)
    {
        RET_ERR(ERASE_ERR);
    }


    RET_OK;
}


// Program stubs which set the device frequency, erase the flash, and program the flash are downloaded
// into ram (by executeCode()). Consequently, it's necessary to preserve the ram before the operation,
// and then to restore the ram after the operation. Since some device contain the RAM CORRUPTED DURING
// JTAG ACCESS bug, it's necesary (safest) to preserve and restore the entire ram, and not just the ram
// used by the stub programs. Unfortunately, this requires extra time.
static CHAR dcoSave[10];
static CHAR IE1Save[4];
//static CHAR IE2Save[2];
STATUS_T preserveRam(void)
{
	const CHAR dummy[2] = {0,0};
	// Save Clock Registers
	if (deviceHasDCOPlus())
	{
		//DCO plus
		if (MSP430_Read_Memory(0x50, dcoSave, 0x54 - 0x50) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	}
	else
	{
		//DCO
		if (MSP430_Read_Memory(0x54, dcoSave, 0x58 - 0x56) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	}
	// Read the IE1/IE2.
	if (MSP430_Read_Memory(bIE1, IE1Save, 2) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	// and clear the IE1/IE2.
	if (MSP430_Write_Memory(bIE1, (char*) dummy, 2) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }

	// Read the ram.
	if (MSP430_Read_Memory(deviceRamStart(), ram, deviceRamEnd() + 1 - deviceRamStart()) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	return (STATUS_OK);
}

STATUS_T restoreRam(void)
{
	int ramSize = deviceRamEnd() + 1 - deviceRamStart();
	int i;

	if (deviceHasDCOPlus())
	{
		//DCO plus
		if (MSP430_Write_Memory(0x50, dcoSave, 0x54 - 0x50) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	}
	else
	{
		//DCO
		if (MSP430_Write_Memory(0x54, dcoSave, 0x58 - 0x56) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	}
	// Write the IE1/IE2.
	if (MSP430_Write_Memory(bIE1, IE1Save, 2) != STATUS_OK)
	    {
		    return (STATUS_ERROR);
	    }
	// Write the ram.
	if (MSP430_Write_Memory(deviceRamStart(), ram, ramSize) != STATUS_OK)
	{
		return (STATUS_ERROR);
	}
	
	// Verify the ram.
	if (MSP430_Read_Memory(deviceRamStart(), verifyRam, ramSize) != STATUS_OK)
	{
		return (STATUS_ERROR);
	}
	for (i = 0; i < ramSize; i++)
	{
		if (verifyRam[i] != ram[i])
		{
			return (STATUS_ERROR); // Verification failure.
		}
	}

	return (STATUS_OK);
}

// Local functions. -----------------------------------------------------------

static STATUS_T setDCOPlus(void) // Set DCO+. MCLK will be divided by 5 within the Flash Controller.
{
	int first = 0, last = 27, mid, reload;
	LONG freq;

	// Binary search through the available frequencies selecting the highest frequency < (476KHz - 10%).
	while (first + 1 < last)
	{
		mid = (last + first) / 2;
		// Select DCO range from 0.23MHz to 11.2MHz. Specify frequency via Ndco. Disable Modulation. Enable DCO+.
		freq = getDCOPlusFreq(0x00, (WORD) (mid << 3), 0x80, 0x80, 0);
		if (freq == 0)
		{
			return (STATUS_ERROR);
		}
		if (freq > 428000 * 5) // Max. Flash Controller frequency - 10%.
		{
			last = mid;
			reload = TRUE;
		}
		else
		{
			first = mid;
			reload = FALSE;
		}
	}

	if (reload)
	{
		freq = getDCOPlusFreq(0x00, (WORD) (first << 3), 0x80, 0x80, 0);
	}
	if ((freq < 257000 * 5) || (freq > 476000 * 5)) // Absolute min and max.
	{
		return (STATUS_ERROR);
	}

	return (STATUS_OK);
}

static LONG getDCOPlusFreq(WORD SCFI0, WORD SCFI1, WORD SCFQCTL, WORD FLL_CTL0, WORD FLL_CTL1)
{
	static WORD code[] = {
		0x0200,               //                 .word   $               ;Load address.
		0x020c,               //                 .word   Main            ;Starting address.
		0x0244,               //                 .word   End             ;Ending address.
		0x0000,               // valSCFI0        .byte   0               ;Dummy value for SCFI0.
		                      // valSCFI1        .byte   0               ;Dummy value for SCFI1.
		0x0000,               // valSCFQCTL      .byte   0               ;Dummy value for SCFQCTL.
		                      // valFLL_CTL0     .byte   0               ;Dummy value for FLL_CTL0.
		0x0000,               // valFLL_CTL1     .byte   0               ;Dummy value for FLL_CTL1.
//		0x40b2,0x5a80,0x0120, // Main            mov     #WDTPW+WDTHOLD,&WDTCTL ;Disable Watchdog.
		0xc232,0xc232,0xc232, //                 ; dummy stop WDT        ;Dummy instructions to replace HoldWDT (should already be stopped)
		0xc232,               //                 dint                    ;Disable interrupts.
//		0xd072,0x0040,        //                 bis.b   #SCG0,SR        ;Disable FLL+.
		0x4032,0x0040,        //                 mov     #SCG0,SR        ;Disable FLL+.
		0x40d2,0xffee,0x0052, //                 mov.b   valSCFQCTL,&SCFQCTL
		0x40d2,0xffe7,0x0051, //                 mov.b   valSCFI1,&SCFI1 ;N(dco).
		0x40d2,0xffe3,0x0053, //                 mov.b   valFLL_CTL0,&FLL_CTL0
		0x40d2,0xffde,0x0054, //                 mov.b   valFLL_CTL1,&FLL_CTL1
		0x40d2,0xffd4,0x0050, //                 mov.b   valSCFI0,&SCFI0 ;D/FN_x.
		0xea0a,               //                 xor     R10,R10         ;Clear R10.
		0xe909,               //                 xor     R9,R9           ;Clear R9.
		0x5319,               // Loop            inc     R9
		0x23fe,               //                 jnz     Loop            ;Increment R9 in a tight loop.
		0x531a,               //                 inc     R10
		0x23fc,               //                 jnz     Loop            ;Increment R10.
		0x4303,               //                 nop
		0x3fff                // End             jmp     $               ;Wait here in case we don't get stopped in time by driver.
	};
	static WORD sizeCode = (WORD) (sizeof(code)/sizeof(*code));
	DWORD startTime;
	DWORD stopTime;
	DWORD elapseTime;

	// Load the parameters. Note the needed byte swap.
	code[3] = (SCFI1 << 8) | SCFI0;
	code[4] = (FLL_CTL0 << 8) | SCFQCTL;
	code[5] = FLL_CTL1;

	// Download the program into memory, verify, and execute.
	if (executeCode(code, sizeCode, executeVerify, FALSE) != STATUS_OK)
	{
		return (0);
	}

	HIL_StartTimer(); // Time the delay (as we could be interrupted, etc.).
	startTime = HIL_ReadTimer(); // System timestamp (in milliseconds).

// Delay time (mSecs).
#define DELAYTIME 100

	HIL_DelayMSec(DELAYTIME); // Delay while program runs in CPU.

	// For F413P, F41xC, and F12x devices, the following step will likely corrupt RAM.
	// However, we don't care about this as the RAM contents will be reloaded in necessary.
	if (syncCPU() != STATUS_OK) // Resume control of the CPU. Read/Write will be set.
	{
		return (0);
	}

	stopTime = HIL_ReadTimer();
	HIL_StopTimer();
	elapseTime = startTime < stopTime ? stopTime - startTime : startTime - stopTime; // System time can wrap.

	// Read counters in R9 and R10.
	HIL_JTAG_IR_(IR_CAPTURE_CPU_REG);
	HIL_JTAG_DR16(9); // Prepare to read R9.
	{
		WORD r9;

		r9 = (WORD) HIL_JTAG_DR16(10); // Read R9 and prepare to read R10.
		// Calculate a device speed (cycles/second). Combine the counter values, multiply by 3 (3 clock cycles per loop),
		// and normalize to one second.
		return ((LONG) ((HIL_JTAG_DR16(10) << 16 | r9) * 3.7 * (1000 / elapseTime)));
	}
}

static STATUS_T setDCO(void) // Set DCO. MCLK will be divided by 5 within the Flash Controller.
{
// WORD DCO = 3; // As an initial speed, select CPU speed slightly > (476KHz - 10%) * 5.
// WORD BCS1 = 7;
  WORD DCO = 5; // As an initial speed, select CPU speed slightly > (476KHz - 10%) * 5.
  WORD BCS1 = 1;
  LONG DCOFreq;
  int i;
  
  getDCOFreq(3, 1);
  return (STATUS_OK);
 
}

static STATUS_T setDCO1(void) // Set DCO. MCLK will be divided by 5 within the Flash Controller.
{
//	WORD DCO = 3; // As an initial speed, select CPU speed slightly > (476KHz - 10%) * 5.
//	WORD BCS1 = 7;
  WORD DCO = 5; // As an initial speed, select CPU speed slightly > (476KHz - 10%) * 5.
  WORD BCS1 = 1;
	LONG DCOFreq;
  int i;

 
	do
	{
		DCOFreq = getDCOFreq(DCO, BCS1);
		if (DCOFreq == 0)
		{
			return (STATUS_ERROR);
		}
//    if (DCOFreq > 450000 * 5) // Check for upper limit - 10%.
		if (DCOFreq > 428000 * 5) // Check for upper limit - 10%.
		{
			if(--DCO <= 0)
			{
				DCO = 7;
				if (--BCS1 <= 0)
				{
					return (STATUS_ERROR); // Couldn't get DCO working with correct frequency.
				}
			}
		}
		else if (DCOFreq < 282000 * 5) // Check for lower limit + 10%.
		{
			if(++DCO > 7)
			{
				DCO = 0;
				if (++BCS1 > 7)
				{
					return (STATUS_ERROR); // Couldn't get DCO working with correct frequency.
				}
			}
		}
		else
		{
			break; // We have a working DCO!
		}
	}
	while (TRUE);

	return (STATUS_OK);
}

LONG getDCOFreq(WORD DCO, WORD BCS1)
{
	static WORD code[] = {
		0x0200,               //                 .word   $               ;Load address.
		0x0208,               //                 .word   Main            ;Starting address.
		0x022e,               //                 .word   End             ;Ending address.
		0x0000,               // valDCO          .byte   0               ;Dummy value for DCO register.
		                      // valBCS1         .byte   0               ;Dummy value for BCS1CTL register.
//		0x40b2,0x5a80,0x0120, // Main            mov     #WDTPW+WDTHOLD,&WDTCTL ;Disable Watchdog.
		0xc232,0xc232,0xc232, //                 ; dummy stop WDT        ;Dummy instructions to replace HoldWDT (should already be stopped)
		0xc232,               //                 dint                    ;Disable interrupts.
		0x403a,0x0206,        //                 mov     #valDCO,R10     ;Load ptr to parameters into R10.
		0x4af2,0x0056,        //                 mov.b   @R10+,&DCOCTL   ;Load DCO value.
		0x4af2,0x0057,        //                 mov.b   @R10+,&BCSCTL1  ;Load BCS1 value.
		0x43c2,0x0058,        //                 mov.b   #0,&BCSCTL2
		0xea0a,               //                 xor     R10,R10         ;Clear R10.
		0xe909,               //                 xor     R9,R9           ;Clear R9.
		0x5319,               // Loop            inc     R9
		0x23fe,               //                 jnz     Loop            ;Increment R9 in a tight loop.
		0x531a,               //                 inc     R10
		0x23fc,               //                 jnz     Loop            ;Increment R10.
		0x4303,               //                 nop
		0x3fff                // End             jmp     $               ;Wait here in case we don't get stopped in time by driver.
	};
	static WORD sizeCode = (WORD) (sizeof(code)/sizeof(*code));
	DWORD startTime;
	DWORD stopTime;
	DWORD elapseTime;

	// Load the DCO and BCS1. Note the needed byte swap (and DCO shift).
	code[3] = (BCS1 << 8) | (DCO <<= 5);

	// Download the program into memory, verify, and execute.
	if (executeCode(code, sizeCode, executeVerify, FALSE) != STATUS_OK)
	{
		return (0);
	}

	HIL_StartTimer(); // Time the delay (as we could be interrupted, etc.).
	startTime = HIL_ReadTimer(); // System timestamp (in milliseconds).

	HIL_DelayMSec(DELAYTIME); // Delay while program runs in CPU.

	// For F413P, F41xC, and F12x devices, the following step will likely corrupt RAM.
	// However, we don't care about this as the RAM contents will be reloaded in necessary.
	if (syncCPU() != STATUS_OK) // Resume control of the CPU. Read/Write will be set.
	{
		return (0);
	}

	stopTime = HIL_ReadTimer();
	HIL_StopTimer();
	elapseTime = startTime < stopTime ? stopTime - startTime : startTime - stopTime; // System time can wrap.

	// Read counters in R9 and R10.
	HIL_JTAG_IR_(IR_CAPTURE_CPU_REG);
	HIL_JTAG_DR16(9); // Prepare to read R9.
	{
		WORD r9;

		r9 = (WORD) HIL_JTAG_DR16(10); // Read R9 and prepare to read R10.
		// Calculate a device speed (cycles/second). Combine the counter values, multiply by 3 (3 clock cycles per loop),
		// and normalize to one second.
		return ((LONG) ((HIL_JTAG_DR16(10) << 16 | r9) * 3.3 * (1000 / elapseTime)));
	}
}

// Assumptions:
// 1) The processor is on an instruction load boundary. This implies that TCLK and R/W are high.
// 2) The program counter does not point into RAM.

static STATUS_T executeCode(WORD* code, WORD sizeCode, BOOL verify, BOOL wait)
{
	unsigned long i;

	// This function downloads programs (set DCO/FLL, erase flash, program flash) into RAM, and starts their execution.
	// F413P, F41xC, and F12x devices contain a silicon bug that can corrupt memory when the JTAG instruction register
	// or the JTAG data register is accessed while TCLK is high. The following sequence of steps has been carefully
	// crafted to work around this bug.
	// The bug destroys the contents of a RAM under the following conditions:
	// - The device is under JTAG control
	// - MCLK/TCLK is high
	// - MAB selects a RAM address
	// - A JTAG instruction is shifted in or a JTAG data is shifted in

	if (sizeCode == 0)
	{
		return (STATUS_OK);
	}
	if (code[0] & 1) // Address must be even.
	{
		return (STATUS_ERROR);
	}
	if (isInstrLoad() != STATUS_OK) // (As an internal check) Must start on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}

	// Set PC for use by DATA_QUICK. HALTing the CPU and preparing to write advances the PC.
	// The PC will be advanced by 2.
	HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
	HIL_TCLK(0);
	//HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16((WORD) (code[0] - 2));
	HIL_TCLK(NEG_EDGE);
	HIL_TCLK(NEG_EDGE);
	HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
	if (HIL_JTAG_DR16(0) != code[0] - 2)
	{
		return (STATUS_ERROR);
	}
	// The next rising TCLK will clock the CPU onto an instruction load boundary.
	
	HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(JMP_D); // Load "jmp $" instruction for use while HALT'ed.
	HIL_TCLK(NEG_EDGE);
	if (isInstrLoad() != STATUS_OK) // (As an internal check) Must be on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}
	cntrlSigLowByte(CNTRL_SIG_HALT_JTAG); // Enable write to memory and halt the CPU (with the next TCLK).
	HIL_JTAG_IR_(IR_DATA_QUICK);
	
	for (i = 0; i < sizeCode; i++)
	{
		HIL_JTAG_DR16(code[i]); // Write data to memory.
		HIL_TCLK(NEG_EDGE);
	}

	cntrlSigLowByte(CNTRL_SIG_READ); // Release the CPU and disable memory write.
	// The next rising TCLK will clock the CPU onto an instruction load boundary.
	
	if (verify) // Verify the memory contents (without corrupting it!)
	{
		WORD *Read = (WORD *) malloc(sizeCode * sizeof(*Read));
		
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
		HIL_TCLK(NEG_EDGE);
		//HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16((WORD) (code[0] - 2));
		HIL_TCLK(NEG_EDGE);
		HIL_TCLK(NEG_EDGE);
		HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
		if (HIL_JTAG_DR16(0) != code[0] - 2)
		{
			free(Read); // Free read buffer
			return (STATUS_ERROR);
		}
		// The next rising TCLK will clock the CPU onto an instruction load boundary.
		
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(JMP_D); // Load "jmp $" instruction for use while HALT'ed.
		HIL_TCLK(NEG_EDGE);
		if (isInstrLoad() != STATUS_OK) // (As an internal check) Must be on an instruction load boundary.
		{
			free(Read); // Free read buffer
			return (STATUS_ERROR);
		}
		// Enable read from memory and halt the CPU (with the next TCLK).
		cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG);
		HIL_JTAG_IR_(IR_DATA_QUICK);
		
		for (i = 0; i < sizeCode; i++)
		{
			HIL_TCLK(NEG_EDGE);
			Read[i] = (WORD) HIL_JTAG_DR16(0); // Read data from memory.
		}
		
		cntrlSigLowByte(CNTRL_SIG_READ); // Release the CPU.
		// The next rising TCLK will clock the CPU onto an instruction load boundary.

		for (i = 0; i < sizeCode; i++) // Verify the memory contents.
		{
			if (Read[i] != code[i])
			{
				free(Read); // Free read buffer
				return (STATUS_ERROR);
			}
		}
		free(Read); // Free read buffer
	} // if (verify)
	
	if (deviceHasJTAGBug())
	{
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(MOV_IMM_PC); // Force PC into non-RAM area.
		HIL_TCLK(NEG_EDGE);
		//HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(ROM_ADDR);
		HIL_TCLK(NEG_EDGE);
		HIL_TCLK(NEG_EDGE);
		HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
		if (HIL_JTAG_DR16(0) != ROM_ADDR)
		{
			return (STATUS_ERROR);
		}
	}
	HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(0x4032); // Clear the Status Register (of possible erroneous Low Power Mode).
	HIL_TCLK(NEG_EDGE);
	//HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(0);
	HIL_TCLK(1);
	HIL_TCLK(POS_EDGE);
	//HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
	HIL_TCLK(NEG_EDGE);
	//HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(code[1]);
	HIL_TCLK(NEG_EDGE);
	HIL_TCLK(NEG_EDGE);
	HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
	if (HIL_JTAG_DR16(0) != code[1])
	{
		return (STATUS_ERROR);
	}
	HIL_JTAG_IR_(IR_CNTRL_SIG_RELEASE); // Release the CPU (but retain JTAG control signals).
	HIL_TCLK(1);

	if (wait)
	{
		// Limit the number of cycles for the code to execute.
		for (i = 0; i < 3000; i++)
		{
			// Poll the device until it is executing the final code instruction.
			if ((readMab() == code[2]) && (readMab() == code[2]) && (readMab() == code[2]))
			{
				// For F413P, F41xC, and F12x devices, the following step will likely corrupt RAM.
				// However, we don't care about this as the RAM contents will be reloaded in necessary.
				return (syncCPU()); // Resume control of the CPU. Read/Write will be set.
			}
            HIL_DelayMSec(1);
		}
		
		return (STATUS_ERROR);
	}

	return (STATUS_OK);
}
