/*========================================================================*\
|                                                                          |
| Jtag.c                                                                   |
|                                                                          |
| This file contains the low-level CPU control functions, and the          |
| low-level JTAG functions.                                                |
|--------------------------------------------------------------------------|
| Project:              MSP430 JTAG interfcae                              |
| Developed using:      MS Visual C++ 5.0                                  |
|--------------------------------------------------------------------------|
| Author:               Dale Wellborn                                      |
| Version:              1.0                                                |
| Initial Version:      20 / 07 / 02                                       |
| Last Change:                                                             |
|--------------------------------------------------------------------------|
| Version history:                                                         |
| Version: 1.03                                                            |
| 25.09.2002 UPSF fixed bug in Verify_PSA (added PUC sequence at func end) |
| 11.10.2002 UPSF Verify PSA allows no odd address and length              |
| 19.12.2002 UPSF Added testConnect_noRelease to avoid unnecessary device  |
|                 releases                                                 |
| Version: 1.1.3                                                           |
| 22.07.2003 UPSF Added STATUS_T IsFuseBlown(void)                         |
| 07.08.2003 UPSF Removed JTAG TAP Reset from testConnect_NoRelease        |
| Version: 1.1.4                                                           |
| 21.11.2003 UPSF test_Connect: Fuse Blown ckeck improved                  |
|                 IsFuseBlown:  Fuse Blown ckeck improved                  |
| Version: 1.1.5                                                           |
|                 Added testConnect_TAPReset                               |
|                 Fixed problem with F123 PSA in RAM                       |
|                                                                          |
| Version: 1.1.5.7                                                         |
| 07.03.2004 UPSF Default settings for mclkCntrl0 are changed in PUC       |
|                                                                          |
|--------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments                                       |
\*========================================================================*/

/*------------------------------------------------------------------------*\
| Remarks:                                                                 |
|                                                                          |
\*------------------------------------------------------------------------*/

// #includes. -----------------------------------------------------------------

#include "Basic_Types.h"
#include "HIL.h"
#include "MSP430_Debug.h"

#include "Globals.h"
#include "JTAG.h"
#include "JTAG_defs.h"
#include "Memory.h"

// Functions. -----------------------------------------------------------------

//----------------------------------------------------------------------------
/* This function compares the computed PSA (Pseudo Signature Analysis) value
   to the PSA value shifted out from target device.
   It is used for very fast data block or erasure verification.
   Arguments: LONG StartAddr (Start address of data block to be checked, must be even)
              LONG Length (number of bytes within data block, must be even)
              CHAR *DataArray (Pointer to array with the data, 0 for Erase Check)
   Result:    LONG (STATUS_OK if comparison was successful, STATUS_ERROR otherwise)
*/
STATUS_T VerifyPSA(LONG StartAddr, LONG Length, CHAR *DataArray)
{
  WORD TDOWORD, i;
  WORD POLY = 0x0805;               // Polynom value for PSA calculation
  WORD PSA_CRC = 0;

  WORD offset = 0;
  CHAR buffer[2];
  

  // Odd address: Verify first byte:  count-- , address++
  if (StartAddr & 1)
  {
     if (readByte((WORD) StartAddr, 1, buffer) != STATUS_OK)
          {
              RET_ERR(READ_MEMORY_ERR);
          }
     if (buffer[0] != DataArray[0])
          RET_ERR(VERIFY_ERR);
     StartAddr++;
     offset++;
     Length--;
  }

  // Odd Length: Verify last byte: count--
  if (Length & 1)
  {
     if (readByte((WORD) (StartAddr+Length-1), 1, buffer) != STATUS_OK)
          {
              RET_ERR(READ_MEMORY_ERR);
          }
     if (buffer[0] != DataArray[Length+offset-1])
          RET_ERR(VERIFY_ERR);
     Length--;
  }


/*
  // Both the start address and length must be even values, and must be within the device memory.
  if ((StartAddr & 1) || (Length & 1) || (StartAddr < 0) || (StartAddr + Length > FLASH_END_ADDR + 1))
  {
      RET_ERR(VERIFY_BOUNDARIES_ERR);
  }
*/
  PSA_CRC = StartAddr-2;       // Start value for PSA calculation
  PUC();          
  holdWatchdog();
  cntrlSigWord(CNTRL_SIG_READ | CNTRL_SIG_TCE1);
  instrLoad();
  HIL_JTAG_IR_(IR_DATA_16BIT);
  HIL_JTAG_DR16(MOV_IMM_PC);
  HIL_TCLK(1);
  HIL_TCLK(0);
  HIL_JTAG_DR16(StartAddr - 2);
  HIL_TCLK(1);
  HIL_TCLK(0);
  HIL_TCLK(1);
  HIL_TCLK(0);  
  HIL_TCLK(1);
  HIL_TCLK(0);  
  HIL_JTAG_IR_(IR_ADDR_CAPTURE);
  HIL_JTAG_DR16(0x0000);
  HIL_JTAG_IR_(IR_DATA_PSA);
  for (i = 0; i < Length; i=i+2)
  {
    // Calculate the PSA (Pseudo Signature Analysis) value  
    if ((PSA_CRC & 0x8000) == 0x8000)
    {
      PSA_CRC ^= POLY;
      PSA_CRC <<= 1;
      PSA_CRC |= 0x0001;
    }
    else
    {
      PSA_CRC <<= 1;
    }
    // if pointer is 0 then use erase check mask, otherwise data 
    &DataArray[0] == 0 ? (PSA_CRC ^= 0xFFFF) : (PSA_CRC ^= (LONG)((BYTE)DataArray[i+offset]+(WORD)(DataArray[i+1+offset]*0x100)));
    
    // Clock through the PSA  
    HIL_TCLK(1);
    HIL_TCLK(0);                    // set here -> Fixes problem with F123 PSA in RAM
    
    HIL_TCK(0);
    HIL_TMS(1);
    HIL_TCK(1);                     // select DR scan
    HIL_TCK(0);
    HIL_TMS(0);
   
    HIL_TCK(1);                     // capture DR
    HIL_TCK(0);
    HIL_TCK(1);                     // shift DR
    HIL_TCK(0);
  
    HIL_TMS(1);
    HIL_TCK(1);                     // exit DR
    HIL_TCK(0);
    
    // PrepTCLK();                  // Set JTAG FSM back into Run-Test/Idle  
    // ClrTCK();                    // No subroutine to make loop fast
    HIL_TCK(1);
    HIL_TMS(0);
    HIL_TCK(0);
    HIL_TCK(1);
    
//    HIL_TCLK(0);
  }
  HIL_JTAG_IR_(IR_SHIFT_OUT_PSA);
  TDOWORD = (WORD) HIL_JTAG_DR16(0x0000);       // read out the PSA value
  HIL_TCLK(1);
  
  // Unfortunately, the exit of the PSA operation seems to leave the CPU in a state
  // where the next write to memory fails. The following sequence was added to work
  // around this problem.
  PUC();          
  holdWatchdog();
  cntrlSigWord(CNTRL_SIG_READ | CNTRL_SIG_TCE1);
  // End of added sequence.

  instrLoad();

  if (TDOWORD == PSA_CRC)
  {
      RET_OK;
  }
  
  RET_ERR(VERIFY_ERR);
}

STATUS_T testConnect(void)
{

//  int i;


  HIL_TDI(1);        // Ensure that TDI is high.
  HIL_TMS(1);       // Reset Jtag state machine.
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE); // Jtag state machine: Test Logic Reset.
  HIL_TCK(0);
  HIL_TMS(0);
  HIL_TCK(1);        // Jtag state machine: Run Test Idle.   
//  HIL_TCK(POS_EDGE); // Jtag state machine: Run Test Idle.

  // This special sequence does an real fuse check again.
  // So that an accidentially set flag that the fuse is blown will
  // be reset.
  HIL_TMS(1);        // 1.25: reset "sense fuse" state again due to: 
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(0);
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(1); 
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(0); 
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(1); 


/*
  for (i = 3; i > 0; i--)   // only one trial could be wrong 
  { 
   HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE); 
   if (DR_Shift16(0xAAAA) == 0x5555) 
    return(STATUS_OK);   // fuse is blown 
  } 
  return(STATUS_ERROR);    // fuse is not blown 
*/


  // The CPU will now be released and executing (as a result of Test-Logic-Reset).
  // Note: The following IR_BYPASS instruction will also release the CPU.
  
  if (HIL_JTAG_IR_(IR_BYPASS) != JTAGVERSION) // Check for valid Jtag version.
  {
     return (STATUS_ERROR); // Invalid version (or no device present).
  }

  return (STATUS_OK);
}


STATUS_T testConnect_TAPReset(void)
{

  HIL_TMS(1);       // Reset Jtag state machine.
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE); // Jtag state machine: Test Logic Reset.
  HIL_TMS(0);
  HIL_TCK(POS_EDGE); // Jtag state machine: Run Test Idle.
  

  // The CPU will NOT be released !!!
  HIL_JTAG_IR_(IR_CNTRL_SIG_16BIT);
  HIL_JTAG_DR16(0x2401);
  if (HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE) != JTAGVERSION)
   {
      return (STATUS_ERROR);
   }

  return (STATUS_OK);
  
}

STATUS_T testConnect_NoRelease(void)
{
/*
  HIL_TMS(1);       // Reset Jtag state machine.
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE);
  HIL_TCK(POS_EDGE); // Jtag state machine: Test Logic Reset.
  HIL_TMS(0);
  HIL_TCK(POS_EDGE); // Jtag state machine: Run Test Idle.
  */

  // The CPU will NOT be released !!!
  HIL_JTAG_IR_(IR_CNTRL_SIG_16BIT);
  HIL_JTAG_DR16(0x2401);
  if (HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE) != JTAGVERSION)
   {
      return (STATUS_ERROR);
   }

  return (STATUS_OK);
  
}


STATUS_T PUC(void) // Power Up Clear.
{
   // Specific changes (EMU_MODE, EMU_FEAT_EN, EMU_CLK_EN) to the General Control register must be followed
   // by a PUC.
   if (deviceClockControl() == GCC_EXTENDED)
   {
      // Assign the user specified Emu Mode to the device (and the controlling global variable)
      // only prior to a PUC. Enable extended Emu Features and Clock Control too.
      genCntrl = emuMode | EMU_FEAT_EN | EMU_CLK_EN;
      if ((jtagCommand(J_INSTR, IR_EMEX_WRITE_CONTROL, NULL)                  || // Configure General Control.
          jtagCommand(J_WORD, (WORD) (genCntrl | CLEAR_STOP | EEM_EN), NULL) ||
          jtagCommand(J_WORD, (WORD) (genCntrl & ~EEM_EN), NULL)               ) != STATUS_OK)
      {
         return (STATUS_ERROR);
      }
      // Configure Module Clock Control register 0 so that the system behavior is similar to the
      // hardwired module clock control (of the F41x). Basically, all timers are stopped and the other
      // modules can finish their current activities.
      if ((jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL)          ||
          jtagCommand(J_WORD, (WORD) (MX_MCLKCNTL0 + MX_WRITE), NULL) ||
//          jtagCommand(J_WORD, 0x60d7, NULL)                        ) != STATUS_OK)
          jtagCommand(J_WORD, mclkCntrl0, NULL)                      ) != STATUS_OK)
      {
         return (STATUS_ERROR);
      }
   }
   
   HIL_TCLK(0);
   cntrlSigWord(CNTRL_SIG_READ | CNTRL_SIG_TCE1 | CNTRL_SIG_PUC); // Assert PUC.
   HIL_TCLK(1);
//   HIL_JTAG_DR16(CNTRL_SIG_READ | CNTRL_SIG_TCE1); // Negate PUC.
   cntrlSigWord(CNTRL_SIG_READ | CNTRL_SIG_TCE1);  // Negate PUC.
 
    if (deviceHasTMRBug()) // Explicitly set TMR.
   {
      HIL_JTAG_DR16(CNTRL_SIG_READ | CNTRL_SIG_TCE1); // Enable access to Flash register.

      HIL_JTAG_IR_(IR_FLASH_16BIT_UPDATE); // Disable Flash Test Mode.
      HIL_JTAG_DR16(FLASH_SESEL1); // Pulse TMR.
      HIL_JTAG_DR16(FLASH_SESEL1 | FLASH_TMR);
      HIL_JTAG_DR16(FLASH_SESEL1);
      HIL_JTAG_DR16(FLASH_SESEL1 | FLASH_TMR); // Set TMR to user mode.

      cntrlSigHighByte(CNTRL_SIG_TCE1); // Disable access to Flash register (Set CNTRL_SIG_TAGFUNCSAT).
   }
   
   HIL_TCLK(POS_EDGE);
   // Capture the PC; it must contain 0xfffe following a PUC.
   if (readMab() != 0xfffe)
   {
      return (STATUS_ERROR);
   }
   HIL_TCLK(POS_EDGE); // Advance to the first instruction.
   HIL_TCLK(POS_EDGE);

    return (isInstrLoad()); // Must exit on an instruction load boundary.
}

STATUS_T holdWatchdog(void)
{
  WORD WDTCTLHold;

//  if (wWDTCTL == 0) // check if WDTCTL should be stored
     if (readWord(WDT, 1, &wWDTCTL) != STATUS_OK)
     {
        return (STATUS_ERROR); // Could not read Watchdog.
     }
  wWDTCTL  = (WDTPW << 8) | (wWDTCTL & 0x00ff); // Combine with password.
  WDTCTLHold = wWDTCTL | WDTHOLD; // Hold the watchdog.
  if (writeWord(WDT, 1, &WDTCTLHold) != STATUS_OK)
  {
     return (STATUS_ERROR); // Could not write Watchdog.
  }

  return (STATUS_OK);
}

/*
STATUS_T setPC(WORD pc) // Set the Program Counter.
{
   if (pc & 1) // PC must be even.
   {
      return (STATUS_ERROR);
   }
   if (cycleStart() != STATUS_OK)
   {
      return (STATUS_ERROR);
   }
   cycle(MOV_IMM_PC); // MOV #<pc>, PC
   cycle(pc);
   if (cycleEnd() != STATUS_OK)
   {
      return (STATUS_ERROR);
   }
   if (readMab() != pc)
   {
      return (STATUS_ERROR); // Couldn't set PC.
   }

   return (isInstrLoad()); // Must exit on an instruction load boundary.
}
*/

STATUS_T setPC(WORD pc) // Set the Program Counter.
{
   if (pc & 1) // PC must be even.
   {
      return (STATUS_ERROR);
   }
   if (cycleStart() != STATUS_OK)
   {
      return (STATUS_ERROR);
   }
   cycle(MOV_IMM_PC); // MOV #<pc>, PC
   cycle(pc);
   if (cycleEnd() != STATUS_OK)
   {
      return (STATUS_ERROR);
   }
   if (readMab() != pc)
   {
      return (STATUS_ERROR); // Couldn't set PC.
   }

   return (isInstrLoad()); // Must exit on an instruction load boundary.
}

STATUS_T cycleStart(void) // CPU must not be HALT'ed.
{
   if (isInstrLoad() != STATUS_OK) // Must be on an instruction load boundary.
   {
      return (STATUS_ERROR);
   }
   // The following statement may be incorrect; the CNTRL_SIG_CPU bit should only be changed
   // when TCLK is low.   
   cntrlSigHighByte(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU); // Enable CPU control.

   return (STATUS_OK);
}

void cycle(WORD data) // Execute a single cycle.
{
   HIL_JTAG_IR_(IR_DATA_16BIT);
   HIL_JTAG_DR16(data);
   HIL_TCLK(POS_EDGE);
}

STATUS_T cycleEnd(void)
{
   HIL_TCLK(0);
   // Restore JTAG control of read/write. Read will be asserted. CPU will be on an instruction load boundary.
   cntrlSigHighByte(CNTRL_SIG_TCE1);
   HIL_TCLK(1);

   return (isInstrLoad());
}

#define MAX_TCLK 10

STATUS_T instrLoad(void)
{
   int i;

   // TCLK is low.
   cntrlSigLowByte(CNTRL_SIG_READ); // Clear CNTRL_SIG_HALT_JTAG.
   HIL_TCLK(1);
   for (i = 0; i < MAX_TCLK; i++) // Clock the CPU until an instruction load. Read/Write is READ.
   {
      if (isInstrLoad() == STATUS_OK)
      {
         break; // Instruction Load is set.
      }
      HIL_TCLK(POS_EDGE); // Otherwise, clock TCLK a maximum of MAX_TCLK times.
   }
   if (i == MAX_TCLK)
   {
      return (STATUS_ERROR);
   }

   return (STATUS_OK); // TCLK is high.
}

STATUS_T isInstrLoad(void) // Is the CPU on an instruction load boundary?
{
   HIL_JTAG_IR_(IR_CNTRL_SIG_CAPTURE);
   if (!(HIL_JTAG_DR16(0) & CNTRL_SIG_INSTRLOAD))
   {
      return (STATUS_ERROR);
   }

   return (STATUS_OK);
}


//------------------------------------------------------------------------
/* This function checks if the security fuse is blown.
   Arguments: None
   Result:     word (STATUS_OK, if fuse is blown, STATUS_ERROR otherwise)
*/
STATUS_T IsFuseBlown(void)
{
    int i;
    WORD cntrlSig;

    testConnect();
    
    for (i = 0; i < 3; i++) // First test could be negative.
    {
        if ((jtagCommand(J_INSTR, IR_CNTRL_SIG_CAPTURE, NULL) ||
             jtagCommand(J_WORD, 0xaaaa, &cntrlSig)             ) != STATUS_OK)
        {
            RET_ERR(BLOW_FUSE_ERR);
        }
        // If the fuse is blown, the device will be in JTAG bypass mode, and data
        // input to the device will be echoed with a one-bit delay.
        if (cntrlSig == 0x5555)
        {
            RET_OK; // Fuse is blown.
        }
    }
    RET_ERR(BLOW_FUSE_ERR); // Fuse is not blown.    
}


WORD readMab(void) // Read Memory Address Bus.
{
   // Note: If this function is used to read the PC, ensure that the CPU is on an instruction load boundary.
   // However, if the CPU is not under JTAG control (i.e., the CPU is not synchronized with the JTAG), this
   // function can be used without requirement of being on an instruction load boundary. Bear in mind that the
   // MAB value will probably not be correct if the CPU is not on an instruction load boundary.
   HIL_JTAG_IR_(IR_ADDR_CAPTURE);
   return ((WORD) HIL_JTAG_DR16(0));
}

WORD cntrlSigWord(WORD data)
{
   data = data & (~CNTRL_SIG_TCE);   // !!!!!!!!!!!!!!!!!!!!!!!!!
    HIL_JTAG_IR_(IR_CNTRL_SIG_16BIT);
   return ((WORD) HIL_JTAG_DR16(data | CNTRL_SIG_TAGFUNCSAT));
}

BYTE cntrlSigHighByte(WORD data)
{
   data = data & (~CNTRL_SIG_TCE);   // !!!!!!!!!!!!!!!!!!!!!!!!!
   HIL_JTAG_IR_(IR_CNTRL_SIG_HIGH_BYTE);
   return ((BYTE) HIL_JTAG_DR8((data | CNTRL_SIG_TAGFUNCSAT) >> 8));
}

BYTE cntrlSigLowByte(BYTE data)
{
   data = data & (~CNTRL_SIG_TCE);   // !!!!!!!!!!!!!!!!!!!!!!!!!
   HIL_JTAG_IR_(IR_CNTRL_SIG_LOW_BYTE);
   return ((BYTE) HIL_JTAG_DR8(data));
}

STATUS_T jtagCommand(BYTE Type, WORD Value, WORD* Result)
{
   LONG result;
   STATUS_T Return;

   switch (Type)
   {
   case J_INSTR:
      {
         result = HIL_JTAG_IR_(Value); // Shift instruction.
         Return = (result == JTAGVERSION) ? STATUS_OK : STATUS_ERROR;
         break;
      }
   case J_WORD:
      {
         result = HIL_JTAG_DR16(Value); // Shift 16-bit data.
         Return = STATUS_OK;
         break;
      }
   case J_BYTE:
      {
         result = HIL_JTAG_DR8(Value); // Shift 8-bit data.
         Return = STATUS_OK;
         break;
      }
   default:
      {
         result = Return = STATUS_ERROR;
         break; // Internal error!
      }
   }
   if (Result != NULL)
   {
      *Result = (WORD) result; // Store result.
   }

   return (Return);
}
