/*========================================================================*\
|                                                                          |
| MSP430.c                                                                 |
|                                                                          |
| This file contains the DLL functions                                     |
|--------------------------------------------------------------------------|
| Project:              MSP430 JTAG interfcae                              |
| Developed using:      MS Visual C++ 5.0                                  |
|--------------------------------------------------------------------------|
| Author:               Dale Wellborn                                      |
| Version:              1.04                                               |
| Initial Version:      20 / 07 / 02                                       |
| Last Change:                                                             |
|--------------------------------------------------------------------------|
| Version history:                                                         |
| Version: 1.0.1                                                           |
| 19.12.2002 UPSF Added testConnect_noRelease to aviod unnecessary device  |
|                 releases                                                 |
| Version: 1.0.2                                                           |
| 20.01.2003 HHOF Activating of breakpoints moved from setBreakpointsActive| 
|                 to MSP430_Breakpoint                                     |
|                                                                          |
| Version: 1.1.0                                                           |
| 27.01.2003 HHOF Breakpoint0 saved before and restored after              |
|                 singlestepping                                           |
|                                                                          |
| Version: 1.1.3                                                           |
| 08.04.2003 UPSF Added detection of FE42x and FW42x devices               |
| 29.04.2003 UPSF Fixed handling of saved ADC10 Registers for FW device    |
| 16.05.2003 UPSF change MSP430_Identify to not release the CPU (Test !!!!)|
| 18.05.2003 UPSF reworked Identification routine so that new devices could|
|                  be added be changing device.c only                      |
| 23.06.2003 UPSF fixed config fuses test used and instead of or           |
| 17.07.2003 UPSF added special handling of I2CDR register                 |
|                 changed structure of MSP430_RESET routine                |
|                 added blown fuse check                                   |
| 24.07.2003 UPSF added update of Registers after PUC to get ADC10 Values  |
|                 updated                                                  |
| 25.07.2003 UPSF added debug Trace function                               |
| 29.07.2003 UPSF added Fuse Pattern for device Identification             |
| 06.08.2003 UPSF added immediate Register write back for all registers    |
|                 except PC & SR -> Registers are updated after Reset      |
| 07.08.2003 UPSF changed MSP430_RESET so that only JTAG_PUC does not      |
|                 relaese CPU - no JTAG State machine Reset is excuted     |
| 18.08.2003 UPSF added RAM starting address to MSP430_Device              |
| 25.08.2003 UPSF MSP430_State does only access the JTAG every 100ms to    |
|                 avoid high CPU load (on WIN32 systems only)              |
| 25.08.2003 UPSF MSP430_Memory fixed problem with write of WDTCTL (WDTPW) |
| Version: 1.1.4                                                           |
| 04.09.2003 UPSF Prevent R3 to been written -> otherwise an error will be |
|                  reported                                                |
| 17.10.2003 UPSF fixed calls of deviceHasI2C to deviceHasI2C()            |
|                                                                          |
| Version: 1.1.5                                                           |
| 11.12.2003 UPSF Prevent R3 to been written -> otherwise an error will be |
| 15.01.2004 UPSF Added FLASH_TEST_MODE in MSP430_Configure Function       |
| 01/20/2004 UPSF Added support for MSP430F161x                            |
| 01/21/2004 UPSF Added ramStart; ram2Start; ram2End; mainEnd; infoEnd to  |
|                 union DEVICE_T                                           |
|                 added deviceInternal union                               |
|                 renamed WDTCTL -> wWDTCTL                                |
|                 moved Flash routines into Flash.c                        |
| 01/22/2004 UPSF Added Delay after reconnect when jtagReleased            |
|                 Added testConnect_TAPReset after reconnect when          |
|                       jtagReleased                                       |
| 02/24/2004 UPSF MSP430_Configure: replaced deviceId by deviceTypeId for  |
|                       correct emulation detection                        |
|                 MSP430_RUN: fixed setting of clock control for single    |
|                       stepping                                           |
| Version: 1.1.5.7                                                         |
| 07.03.2004 UPSF Added mclkCntrl0 variable in MSP430_Identify and         |
|                     MSP430_Configure                                     |
|                 Added filtering and saving of Clock settings in          |
|                     _msp430_EEM_Write_Register                           |
|                                                                          |
|--------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments                                       |
\*========================================================================*/

/*------------------------------------------------------------------------*\
| Remarks:                                                                 |
|                                                                          |
\*------------------------------------------------------------------------*/

// #includes. -----------------------------------------------------------------

#include <stdlib.h>
#include <string.h>

#include "MSP/MSP430_Debug.h"

#include "MSP/Cycles.h"     // Cycle counter
#include "MSP/Execution.h"  // Execution functions.
#include "MSP/Flash.h"      // Flash functions.
#include "MSP/Globals.h"    // Global variables and HIL (Hardware Interface Layer) macros/functions.
#include "MSP/JTAG.h"
#include "MSP/JTAG_defs.h"  // JTAG instruction and bit definitions.
#include "MSP/Memory.h"     // Memory read/write.

// #defines. ------------------------------------------------------------------


#define DeviceIDAddress0         0x0
#define DeviceIDAddress1         0x1
#define RevisionAddress          0x2
#define FabIDAddress             0x3
#define SelfTestCodeAddress0     0x8
#define SelfTestCodeAddress1     0x9
#define EmbeddedSystemIDAddress  0xD

#define DLL_VERSION         10105   // DLL version * 1.01.05.
#define SLOWFUSECHECK_BUG
#define DEFAULT_VCC         3000    // Default Vcc to 3V (3000mV).
#define MAX_VCC             3600    // 3600mV (3.6V) maximum.
#define N_REGS              16

// (File) Global variables. ---------------------------------------------------

static BOOL jtagReleased = FALSE;
static int runMode = STOPPED;
static WORD bpState;
static BOOL extraStep = FALSE;

union DEVICE_T deviceInternal; // Device information structure.


#ifndef uController
static const char* errorStrings[] =
{
    "No error",
    "Could not initialize device interface",
    "Could not close device interface",
    "Invalid parameter(s)",
    "Could not find device (or device not supported)",
    "Unknown device",
    "Could not read device memory",
    "Could not write device memory",
    "Could not read device configuration fuses",
    "Incorrectly configured device; device derivative not supported",
    "Could not set device Vcc",
    "Could not reset device",
    "Could not preserve/restore device memory",
    "Could not set device operating frequency",
    "Could not erase device memory",
    "Could not set device breakpoint",
    "Could not single step device",
    "Could not run device (to breakpoint)",
    "Could not determine device state",
    "Could not open Enhanced Emulation Module",
    "Could not read Enhanced Emulation Module register",
    "Could not write Enhanced Emulation Module register",
    "Could not close Enhanced Emulation Module",
    "File open error",
    "Could not determine file type",
    "Unexpected end of file encountered",
    "File input/output error",
    "File data error",
    "Verification error",
    "Could not blow device security fuse",
    "Could not access device - security fuse is blown",
    "Error within Intel Hex file",
    "Could not write device Register",
    "Invalid error number",
};
#endif

// Global variables. ----------------------------------------------------------

int errorNumber = NO_ERR;       // Error number.
int deviceId = DEVICE_UNKNOWN;  // Device identification.
BOOL hardwareBPsUsed = FALSE;   // Hardware breakpoints are used.
LONG deviceRegs[N_REGS];        // Device registers.
WORD savedAdc10dtc0;            // ADC10 registers.
WORD savedAdc10ctl0;
WORD savedAdc10ctl1;
WORD wWDTCTL = 0;               // Watchdog Timer Control register.
WORD emuMode = EMU_MODE_F44X_100; // 4xx emulation mode.
WORD genCntrl = 0;              // General Control register.
WORD clkCntrl = 0;              // Clock Control register.
// Configure Module Clock Control register 0 so that the system behavior is similar to the
// hardwired module clock control (of the F41x). Basically, all timers are stopped and the other
// modules can finish their current activities.
#define DEFAULT_MCLKCNTRL0    0x60d7
WORD mclkCntrl0 = DEFAULT_MCLKCNTRL0;// Module Clock Control register.

BOOL executeVerify = FALSE;     // Verify the stub code downloaded by executeCode().
BOOL eemOpen = FALSE;           // Enhanced Emulation Mode is open.
LONG CPUCycles = -1;            // CPU cycles executed during single step.

struct
{
   LONG value;
    LONG mask;
    LONG cntrl;
   LONG comb;
   LONG cpuStop;
   BYTE valid;
} brkPnt0Bckup = {0,0,0,0,0,0}; // Backup of breakpoint while single stepping


#ifdef DEBUG_TRACE
  #include <stdio.h>
  #include <time.h>    
  ULONG timeDiff = 0;
  ULONG lastTime = 0;
  FILE *tracefile;
#endif


// Function Declarations. -----------------------------------------------------

STATUS_T _msp430_Identify(CHAR* buffer, LONG count, LONG setId);


// Functions. -----------------------------------------------------------------

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Initialize(CHAR* port, LONG* version);

Description:
 Initialize the interface.

Parameters:
 port:    Interface port reference (application specific).
 version: The version number of the MSP430 DLL is returned.

Returns:
 STATUS_OK:    The interface was initialized.
 STATUS_ERROR: The interface was not initialized.

Error codes:
 INITIALIZE_ERR

Notes:
 1. *** This function must be called first. ***
 2. *** MSP430_Vcc() must be called second (after MSP430_Initialize() is called). ***
 3. This function calls user-supplied function: HIL_Initialize(port)
*/
STATUS_T WINAPI MSP430_Initialize(CHAR* port, LONG* version)
{
#ifdef DEBUG_TRACE
    HIL_Initialize("traceon");
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "************************************************************\n");
    fprintf(tracefile, "*** New Session                                          ***\n");
    fprintf(tracefile, "************************************************************\n");
    fprintf(tracefile, "%5d msec MSP430_Initialize Port: 0x%4.4X  \n",timeDiff, *port);
      fclose(tracefile);
#endif
    
    if (version)
    {
        *version = DLL_VERSION;
    }

    if ((HIL_Initialize(port)) != STATUS_OK)
    {
        RET_ERR(INITIALIZE_ERR);
    }

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Close(LONG vccOff);

Description:
 Close the interface.

Parameters:
 vccOff: Turn off the device Vcc (0 volts) if TRUE.

Returns:
 STATUS_OK:    The interface was closed.
 STATUS_ERROR: The interface was not closed.

Error codes:
 CLOSE_ERR

Notes:
 1. If called, this function must be called last.
 2. This function calls user-supplied function: HIL_Close()
*/
STATUS_T WINAPI MSP430_Close(LONG vccOff)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Close vccOff: 0x%4.4X  \n",timeDiff, vccOff);
      fclose(tracefile);
#endif
    if (HIL_Close(vccOff) != STATUS_OK)
    {
        RET_ERR(CLOSE_ERR);
    }

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Identify(CHAR* buffer, LONG count, LONG setId);

Description:
 Identify the device, and obtain device information.

Parameters:
 buffer: A pointer to a buffer where the device identity and information is stored. The values are returned
         in the buffer at the following offsets:
          0: The value 0xaa55 [so one can determine how to unpack the subsequent fields] (WORD).
          2: Device identification (WORD). Refer to the "Devices" enumeration.
          4: Identification string (CHAR[32])
         36: MAIN MEMORY (FLASH) starting address (WORD). Ending address: 0xffff
         38: INFORMATION MEMORY (FLASH) starting address (WORD). Ending address: 0x10ff
         40: RAM ending address (WORD). Starting address: 0x0200
         42: Number of breakpoints (WORD)
         44: Emulation level (WORD)
         46: Clock control level (WORD)
         48: LCD starting address (WORD)
         50: LCD ending address (WORD)
         52: Vcc minimum during operation [mVolts] (WORD)
         54: Vcc maximum during operation [mVolts] (WORD)
         56: Device has TEST/VPP (WORD)
         58: RAM starting address (WORD).
         60: RAM2 starting address (WORD).
         62: RAM2 ending address (WORD).
         64: INFO ending address (WORD).
         66: Main ending address (LONG).
 count: The number of bytes to return in the buffer (i.e., the buffer size).
 setId: If setId is not DEVICE_UNKNOWN, the device is set to setId. Otherwise the device is determined.

Returns:
 STATUS_OK:    The device identify and information was obtained.
 STATUS_ERROR: The device identity and information was not obtained.

Error codes:
 PARAMETER_ERR
 NO_DEVICE_ERR
 READ_MEMORY_ERR
 READ_FUSES_ERR
 CONFIGURATION_ERR

Notes:
 1. It is possible to successfully (i.e. return status: STATUS_OK) identify an "UNKNOWN DEVICE".
 2. Use setId and MSP430_State(state, TRUE, FALSE) to stop a running device (without resetting it).
 3. Following MSP430_Identify():
    - the device is reset, is not executing, and is under JTAG control.
    - the JTAG interface is enabled and the JTAG signals are negated.
    - Vpp is set to the level specified in HIL_Open().
*/
STATUS_T WINAPI MSP430_Identify(CHAR* buffer, LONG count, LONG setId)
{
  union DEVICE_T* device = (union DEVICE_T*) buffer; // Device information structure.
  STATUS_T status;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Identify setId: 0x%4.4X  \n",timeDiff, setId);
      fclose(tracefile);
#endif


  if ((!buffer) || (count < 0) || (setId < DEVICE_UNKNOWN))
  {
      RET_ERR(PARAMETER_ERR);
  }

  status = _msp430_Identify(buffer, count, setId);

  if (sizeof(deviceInternal) < count) count = sizeof(deviceInternal);
  memcpy(&device->endian, &deviceInternal.endian, count);

  return(status);
}


STATUS_T _msp430_Identify(CHAR* buffer, LONG count, LONG setId)
{
  BYTE iddata[18];

  if ((!buffer) || (count < 0) || (setId < DEVICE_UNKNOWN))
  {
      RET_ERR(PARAMETER_ERR);
  }

  // (Re-) Initialize the File Global variables.
  eraseFrequencySet = FALSE;
  programFrequencySet = FALSE;
  jtagReleased = FALSE;
  runMode = STOPPED;
  extraStep = FALSE;

  // (Re-) Initialize the Global variables.
  errorNumber = NO_ERR;
  deviceId = DEVICE_UNKNOWN;
  deviceTypeId = DEVICE_UNKNOWN; // Clear the device derivative in table.
  hardwareBPsUsed = FALSE;
  emuMode = EMU_MODE_F44X_100;
  genCntrl = 0;
  clkCntrl = 0;
  mclkCntrl0 = DEFAULT_MCLKCNTRL0; // Module Clock Control register.
  executeVerify = FALSE;
  eemOpen = FALSE;
  CPUCycles = -1;

  // (Re-) Initialize the default device Settings
  deviceInternal.ramStart = 0x0200;
  deviceInternal.ram2Start = 0x0000;
  deviceInternal.ram2End = 0x0000;
  deviceInternal.infoEnd = 0x10FF;
  deviceInternal.mainEnd = 0xFFFF;
  deviceInternal.mainStart = 0x1100;
  deviceInternal.infoStart = 0x1000;
  deviceInternal.ramEnd = 0x9FF;

  HIL_Open(); // Enable the JTAG interface to the device.
  // Check the fuse.
#ifdef SLOWFUSECHECK_BUG // Silicon bug requires slow check of fuse.
  HIL_TCK(0);
  HIL_TMS(0);
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(1);
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(0);
  HIL_DelayMSec(FUSECHECK_DELAY);
  HIL_TMS(1);
  HIL_DelayMSec(FUSECHECK_DELAY);
#else
  HIL_TMS(0);
  HIL_TMS(1);
  HIL_TMS(0);
  HIL_TMS(1);
#endif

  if (setId) // Explicitly set the device id, and initialize the software state in anticipation
             // of stopping a running device (without reset).
  {
      //set deviceID and deviceTypeId
      deviceId = setId;
      deviceTypeId = sdeviceID[deviceId].deviceTypeID;

      eraseFrequencySet = FALSE;
      programFrequencySet = FALSE;
      jtagReleased = FALSE;
      runMode = RUNNING;

      if (testConnect() != STATUS_OK)
      {
          if (IsFuseBlown() == STATUS_OK) RET_ERR(FUSE_BLOWN_ERR);

          RET_ERR(NO_DEVICE_ERR); // Could not find device (or device not supported).
      }
  }
  else
  {
     
     do
      {
        int i=0;
          
          deviceId = DEVICE_UNKNOWN; // Clear the device derivative.
          deviceTypeId = DEVICE_UNKNOWN; // Clear the device derivative in table.
          // Connect to, synchronize with, and reset the CPU. The CPU is clocked with TCLK. Read/Write will be set.
          if (MSP430_Reset(ALL_RESETS, FALSE, FALSE) != STATUS_OK)
          {
              RET_ERR(errorNumber); // Could not find device (or device not supported).
          }
          if (MSP430_Read_Memory(0x0ff0, iddata, 16) != STATUS_OK)
          {
              RET_ERR(READ_MEMORY_ERR); // Could not read device memory.
          }

        while (i <= NUMofDevices)
        {
              // sdeviceID[i].deviceIDParameter[NDEVICEIDPARAMETER];
              //0x0   0x1   0x2   0x3   0x8   0x9   0xD   Fuse
              //ID0   ID1   Rev   Fab   Self0 Self1 Config
           if (\
              ((sdeviceID[i].deviceIDParameter[0] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[0] == iddata[DeviceIDAddress0])       ) &&\
              ((sdeviceID[i].deviceIDParameter[1] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[1] == iddata[DeviceIDAddress1])       ) &&\
              ((sdeviceID[i].deviceIDParameter[2] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[2] == iddata[RevisionAddress])        ) &&\
              ((sdeviceID[i].deviceIDParameter[3] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[3] == iddata[FabIDAddress])           ) &&\
              ((sdeviceID[i].deviceIDParameter[4] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[4] == iddata[SelfTestCodeAddress0])   ) &&\
              ((sdeviceID[i].deviceIDParameter[5] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[5] == iddata[SelfTestCodeAddress1])   ) &&\
              ((sdeviceID[i].deviceIDParameter[6] == (short)(-1)) || (sdeviceID[i].deviceIDParameter[6] == iddata[EmbeddedSystemIDAddress])) \
              )
           {
              // Test for config Fuses and check configuration
              if ((sdeviceID[i].deviceIDParameter[7] != (short)(-1)))
              {
                 WORD ConfigFuses;
                 if ((jtagCommand(J_INSTR, IR_CONFIG_FUSES, NULL) || // Read configuration fuses.
                    jtagCommand(J_BYTE, 0x0000, &ConfigFuses)     ) != STATUS_OK)
                 {
                    RET_ERR(READ_FUSES_ERR); // Could not determine configuration ofthe device derivative.
                 }
                 if ((sdeviceID[i].deviceIDParameter[7] & sdeviceID[i].deviceIDParameter[8]) == (ConfigFuses & sdeviceID[i].deviceIDParameter[8])) // Fuse Lock Bit is added to compare
                 {   
                    deviceId = i; // set the device derivative in table.
                    deviceTypeId = sdeviceID[deviceId].deviceTypeID;
                    break;
                 }
              }
              else
              {
                  deviceId = i; // set the device derivative in table.
                  deviceTypeId = sdeviceID[deviceId].deviceTypeID;
                  break;
              }
           }
           i++;
          } //while (i <= NUMofDevices)

      } while (0); // end do
  } //  if (setId)
  
  // Fill-in the device information structure (if there's adequate space).
  deviceInternal.endian = 0xaa55;
  deviceInternal.id = deviceId;
  strcpy((char *)deviceInternal.string, deviceString());
  deviceInternal.mainStart = deviceMainStart();
  deviceInternal.infoStart = deviceInfoStart();
  deviceInternal.ramEnd = deviceRamEnd();
  deviceInternal.nBreakpoints = deviceNBreakpoints();
  deviceInternal.emulation = deviceEmulation();
  deviceInternal.clockControl = deviceClockControl();
  deviceInternal.lcdStart = deviceLcdStart();
  deviceInternal.lcdEnd = deviceLcdEnd();
  deviceInternal.vccMinOp = deviceVccMinOp();
  deviceInternal.vccMaxOp = deviceVccMaxOp();
  deviceInternal.hasTestVpp = deviceHasTestVpp();
  deviceInternal.ramStart = deviceRamStart();
  deviceInternal.ram2Start = deviceRam2Start();
  deviceInternal.ram2End = deviceRam2End();
  deviceInternal.infoEnd = deviceInfoEnd();
  deviceInternal.mainEnd = deviceMainEnd();


  RET_OK;
}


/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Device(LONG deviceId, CHAR* buffer, LONG count);

Description:
 Obtain the device information.

Parameters:
 deviceId: An index (from zero to DEVICE_END) used to specify the device.
 buffer: A pointer to a buffer where the device identity and information is stored. The values are returned
         in the buffer at the following offsets:
          0: The value 0xaa55 [so one can determine how to unpack the subsequent fields] (WORD).
          2: Device identification (WORD). Refer to the "Devices" enumeration
          4: Identification string (CHAR[32])
         36: MAIN MEMORY (FLASH) starting address (WORD). Ending address: 0xffff
         38: INFORMATION MEMORY (FLASH) starting address (WORD). Ending address: 0x10ff
         40: RAM ending address (WORD). 
         42: Number of breakpoints (WORD)
         44: Emulation level (WORD)
         46: Clock control level (WORD)
         48: LCD starting address (WORD)
         50: LCD ending address (WORD)
         52: Vcc minimum during operation [mVolts] (WORD)
         54: Vcc maximum during operation [mVolts] (WORD)
         56: Device has TEST/VPP (WORD)
         58: RAM starting address (WORD).
         60: RAM2 starting address (WORD).
         62: RAM2 ending address (WORD).
         64: INFO ending address (WORD).
         66: Main ending address (LONG).
    count: The number of bytes to return in the buffer (i.e., the buffer size).

Returns:
 STATUS_OK:    The device identify and information was obtained.
 STATUS_ERROR: The device identity and information was not obtained.

Error codes:
 PARAMETER_ERR

Notes:
 1. This function can be used to determine the devices supported by the driver; a device
    identification of DEVICE_END is returned to indicate that there are no additional devices.
*/
STATUS_T WINAPI MSP430_Device(LONG localDeviceId, CHAR* buffer, LONG count)
{
  union DEVICE_T* device = (union DEVICE_T*) buffer; // Device information structure.
  int globalDeviceId = deviceId; // Preserve the (global) device identification.

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Device localDeviceId: 0x%4.4X  \n",timeDiff, MSP430_Device);
      fclose(tracefile);
#endif


  if ((localDeviceId < 0) || (localDeviceId >= NUMofDevices) || (!buffer) || (count < 0))
//  if ((localDeviceId < 0) || (localDeviceId > DEVICE_END) || (!buffer) || (count < 0))
  {
      RET_ERR(PARAMETER_ERR);
  }

  if (sizeof(deviceInternal) < count) count = sizeof(deviceInternal);

  deviceId = localDeviceId; // Set the (global) device identification (for device() functions).
  deviceTypeId = sdeviceID[deviceId].deviceTypeID;

  // Fill-in the device information structure (if there's adequate space).
  if (count >= 2) device->endian = 0xaa55;
  if (count >= 4) device->id = (WORD) deviceId;
//  if (count >= 36) strcpy(device->string, deviceString());
  if (count >= 36) strcpy((char *)device->string, (char const *)deviceString());
  if (count >= 38) device->mainStart = deviceMainStart();
  if (count >= 40) device->infoStart = deviceInfoStart();
  if (count >= 42) device->ramEnd = deviceRamEnd();
  if (count >= 44) device->nBreakpoints = deviceNBreakpoints();
  if (count >= 46) device->emulation = deviceEmulation();
  if (count >= 48) device->clockControl = deviceClockControl();
  if (count >= 50) device->lcdStart = deviceLcdStart();
  if (count >= 52) device->lcdEnd = deviceLcdEnd();
  if (count >= 54) device->vccMinOp = deviceVccMinOp();
  if (count >= 56) device->vccMaxOp = deviceVccMaxOp();
  if (count >= 58) device->hasTestVpp = deviceHasTestVpp();
  if (count >= 60) device->ramStart = deviceRamStart();

  if (count >= 62) device->ram2Start = deviceRam2Start();
  if (count >= 64) device->ram2End = deviceRam2End();
  if (count >= 66) device->infoEnd = deviceInfoEnd();
  if (count >= 70) device->mainEnd = deviceMainEnd();

  deviceId = globalDeviceId; // Restore the (global) device identification.
  deviceTypeId = sdeviceID[deviceId].deviceTypeID;

  RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Configure(LONG mode, LONG value);

Description:
 Configure the mode(s) of the device and/or the software.

Parameters:
 mode:     The mode to be configured.
 value:    The mode value.

Returns:
 STATUS_OK:    The mode was configured.
 STATUS_ERROR: The mode was not configured.

Error codes:
 PARAMETER_ERR

Notes:
 1. Specify a mode of VERIFICATION_MODE to verify data downloaded into the device FLASH memories.
*/
STATUS_T WINAPI MSP430_Configure(LONG mode, LONG value)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Configure mode: 0x%4.4X   value: 0x%4.4X  \n",timeDiff, mode, value);
      fclose(tracefile);
#endif

    switch (mode)
    {
    case VERIFICATION_MODE:
        executeVerify = (BOOL) value;
        break;
    case EMULATION_MODE:
        if (deviceTypeId != DT_MSP430F44x ) //DEVICE_MSP430F44x)
        { 
            RET_ERR(PARAMETER_ERR);
        }
        emuMode = (WORD) value & (EMU_MODE_F44X_100 | EMU_MODE_F43X_100 | EMU_MODE_F4XX_80 | EMU_MODE_F4XX_64);
        
        // The F449 must be PUC'ed to effect changes to the device emulation.
        return (MSP430_Reset(PUC_RESET, FALSE, FALSE));
        //break;
    case CLK_CNTRL_MODE:
        clkCntrl = (WORD) value;
        break;
    case FLASH_TEST_MODE:
        {
            STATUS_T status;
            //if ((deviceId  & 0xA000) !=  0xA000)  RET_ERR(PARAMETER_ERR);
         if (! deviceHasMarginalReadFlash()) RET_ERR(PARAMETER_ERR);
            status = setupMarginalReadFlash(value);
            if (status != STATUS_OK) return (status);
        }
    default: 
        RET_ERR(PARAMETER_ERR);
        //break;

    }

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_VCC(LONG voltage);

Description:
 Set the device Vcc pin to voltage/1000 volts.

 *** The sample HIL.dll that interfaces to the MSP430 via the FET Interface Module connected
 *** to the parallel port of the PC >>> DOES NOT <<< implement the functionality required to
 *** set the device Vcc pin to voltage/1000 volts.

Parameters:
 voltage: The device Vcc pin is set to voltage/1000 volts.

Returns:
 STATUS_OK:    The Vcc was set to voltage.
 STATUS_ERROR: The Vcc was not set to voltage.

Error codes:
 PARAMETER_ERR
 VCC_ERR

Notes:
 1. This function calls user-supplied function: HIL_VCC(voltage).
 2. A "voltage" of zero (0) turns off voltage to the device.
 3. *** This function must be called second (after MSP430_Initialize() is called). ***
*/
STATUS_T WINAPI MSP430_VCC(LONG voltage)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_VCC voltage: 0x%4.4X  \n",timeDiff, voltage);
      fclose(tracefile);
#endif

    if ((voltage < 0) || (voltage > MAX_VCC))
    {
        RET_ERR(PARAMETER_ERR);
    }

    if (HIL_VCC(voltage) != STATUS_OK)
    {
        RET_ERR(VCC_ERR);
    }

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Reset(LONG method, LONG execute, LONG releaseJTAG);

Description:
 Reset the device using the specified method(s). Optionally start device execution,
 and release the JTAG control signals.

Parameters:
 method:      The bit mask specifying the method(s) to use to reset the device:
                PUC:     The device is reset using PUC (i.e., a "soft" reset).
                RST_NMI: The device is reset using the RST/NMI pin (i.e., a "hard" reset).
                VCC:     The device is reset by cycling power to the device.
 execute:     Start device execution (when TRUE).
 releaseJTAG: Release the JTAG control signals (when TRUE). execute must be TRUE.

Returns:
 STATUS_OK:    The device was reset (and optionally started [and JTAG control released])
 STATUS_ERROR: The device was not reset (and optionally started [and JTAG control released]).

Error codes:
 PARAMETER_ERR
 RESET_ERR
 RUN_ERR
 NO_DEVICE_ERR

Notes:
 1. It is possible to combine reset methods. The methods are applied in the following order:
    PUC then RST_NMI then VCC. If a reset operation fails, the next reset method (if any) is applied.
 2. Following reset by RST/NMI and/or VCC, a PUC is automatically executed to reset the device in
    such a way that it does not begin to execute a resident program (or garbage).
 3. The device registers are updated (for read using MSP430_Registers()) following MSP430_Reset().
*/
STATUS_T WINAPI MSP430_Reset(LONG method, LONG execute, LONG releaseJTAG)
{
    int i = 0;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Reset method: 0x%4.4X   execute: 0x%4.4X   release: 0x%4.4X  JTAG\n",timeDiff, method, execute, releaseJTAG);
      fclose(tracefile);
#endif

    if (!method)
    {
        RET_ERR(PARAMETER_ERR);
    }

    CPUCycles = -1; // Reset count of CPU cycles executed.

    // Force a frequency set after reset.
    eraseFrequencySet = programFrequencySet = FALSE;


    if (1) // (method & PUC_RESET)
    {
        
        for (i=0;i < 10;i++)
        {
            // Connect and synchronize with the CPU. The CPU is clocked with TCLK. Read/Write will be set.
            if (testConnect_NoRelease() != STATUS_OK)
              if (syncCPU() != STATUS_OK)
		continue;

                if (PUC() == STATUS_OK)
                {
                    if (!execute)
                    {
                        if (readAllDeviceRegisters(deviceRegs, TRUE) != STATUS_OK)
                        {
                            RET_ERR(RESET_ERR);
                        }
                        runMode = STOPPED;
                    }
                    else
                    {
                        if (releaseCPU() != STATUS_OK) // Start device execution (and retain JTAG control signals).
                        {
                            RET_ERR(RUN_ERR);
                        }
                        runMode = FREE_RUN;

                        if (releaseJTAG)
                        {
                            if (HIL_Release() != STATUS_OK) // Release the JTAG control signals.
                            {
                                RET_ERR(RUN_ERR);
                            }
                            jtagReleased = TRUE;
                        }
                    }
                    RET_OK;
                }
		
         }// for
    } // PUC



    if (method & RST_RESET)
    {
        // Assert (hard) ~RST/NMI.
        HIL_RST(0);
        HIL_RST(1);
        // Check the fuse.
#ifdef SLOWFUSECHECK_BUG // Silicon bug requires slow check of fuse.
        HIL_TCK(0);
        HIL_TMS(0);
        HIL_DelayMSec(FUSECHECK_DELAY);
        HIL_TMS(1);
        HIL_DelayMSec(FUSECHECK_DELAY);
        HIL_TMS(0);
        HIL_DelayMSec(FUSECHECK_DELAY);
        HIL_TMS(1);
        HIL_DelayMSec(FUSECHECK_DELAY);
#else
        HIL_TMS(0);
        HIL_TMS(1);
        HIL_TMS(0);
        HIL_TMS(1);
#endif
        for (i=0;i < 10;i++)
        {
            // Connect and synchronize with the CPU. The CPU is clocked with TCLK. Read/Write will be set.
            if ((testConnect() || syncCPU()) == STATUS_OK)
            {
                if (PUC() == STATUS_OK)
                {
                    if (!execute)
                    {
                        if (readAllDeviceRegisters(deviceRegs, TRUE) != STATUS_OK)
                        {
                            RET_ERR(RESET_ERR);
                        }
                        runMode = STOPPED;
                    }
                    else
                    {
                        if (releaseCPU() != STATUS_OK) // Start device execution (and retain JTAG control signals).
                        {
                            RET_ERR(RUN_ERR);
                        }
                        runMode = FREE_RUN;

                        if (releaseJTAG)
                        {
                            if (HIL_Release() != STATUS_OK) // Release the JTAG control signals.
                            {
                                RET_ERR(RUN_ERR);
                            }
                            jtagReleased = TRUE;
                        }
                    }
                    RET_OK;
                }
            } // if
        } // for
    } // RST/NMI

    if (method & VCC_RESET)
    {
        HIL_VCC(0);
        HIL_DelayMSec(5000); // Delay 5 seconds to give the device time to power-down,
        HIL_VCC(DEFAULT_VCC);;
        HIL_DelayMSec(10); // Delay 10mSec to give the device time to power-up.
        // Assert (hard) ~RST/NMI.
        HIL_RST(0);
        HIL_RST(1);
        // Check the fuse.
#ifdef SLOWFUSECHECK_BUG // Silicon bug requires slow check of fuse.
        HIL_TCK(0);
        HIL_TMS(0);
        HIL_DelayMSec(FUSECHECK_DELAY);
        HIL_TMS(1);
        HIL_DelayMSec(FUSECHECK_DELAY);
        HIL_TMS(0);
        HIL_DelayMSec(FUSECHECK_DELAY);
        HIL_TMS(1);
        HIL_DelayMSec(FUSECHECK_DELAY);
#else
        HIL_TMS(0);
        HIL_TMS(1);
        HIL_TMS(0);
        HIL_TMS(1);
#endif
        for (i=0;i < 10;i++)
        {
            // Connect and synchronize with the CPU. The CPU is clocked with TCLK. Read/Write will be set.
            if ((testConnect() || syncCPU()) == STATUS_OK)
            {
                if (PUC() == STATUS_OK)
                {
                    if (!execute)
                    {
                        if (readAllDeviceRegisters(deviceRegs, TRUE) != STATUS_OK)
                        {
                            RET_ERR(RESET_ERR);
                        }
                        runMode = STOPPED;
                    }
                    else
                    {
                        if (releaseCPU() != STATUS_OK) // Start device execution (and retain JTAG control signals).
                        {
                            RET_ERR(RUN_ERR);
                        }
                        runMode = FREE_RUN;

                        if (releaseJTAG)
                        {
                            if (HIL_Release() != STATUS_OK) // Release the JTAG control signals.
                            {
                                RET_ERR(RUN_ERR);
                            }
                            jtagReleased = TRUE;
                        }
                    }
                    RET_OK;
                }
            } // if
        } // for
    } // VCC Reset

    if (IsFuseBlown() == STATUS_OK) RET_ERR(FUSE_BLOWN_ERR);
    if (testConnect_NoRelease() != STATUS_OK)   // Test !!!!!!!!!!!!!!11
    {
        RET_ERR(NO_DEVICE_ERR); // Could not find device (or device not supported).
    }

    if (method & VCC_RESET)
    {
        RET_ERR(NO_DEVICE_ERR);
    }
    else
    {
        RET_ERR(RESET_ERR);
    }
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Registers(LONG registers[], LONG mask, LONG rw);

Description:
 Read and write the device registers.

Parameters:
 registers: The destination of registers read from the device (rw = READ),
            and the source of registers written to the device (rw = WRITE).
 mask:      A bit-mask which identifies which registers are read/written
            (Bit 0: R0, Bit 1: R1, and so on).
 rw:        Specify a read (READ) or write (WRITE) operation.

Returns:
 STATUS_OK:    The registers were read or written.
 STATUS_ERROR: The registers were not read or written.

Error codes:
 PARAMETER_ERR

Notes:
 1. It is not possible to read and write registers CG1 and CG2.
 2. This function does not read and write the actual device registers. Rather, copies of the device
    registers are read and written. The register copies are updated (for read) after a device reset,
    and after MSP430_State() indicates BREAKPOINT_HIT, SINGLE_STEP_COMPLETE, or STOPPED (after
    stopping a running device [stop = TRUE]). The register copies are written to the device prior to
    MSP430_Run() with FREE_RUN, SINGLE_STEP, and RUN_TO_BREAKPOINT.
*/
STATUS_T WINAPI MSP430_Registers(LONG registers[], LONG mask, LONG rw)
{
    int i;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Registers rw: 0x%4.4X  \n",timeDiff, rw);
      fclose(tracefile);
#endif

    if (!registers || !mask)
    {
        RET_ERR(PARAMETER_ERR);
    }
    
    for (i = 0; i < N_REGS; i++)
    {
        if (mask & MASKREG(i))
        {
            if (rw == READ)
            {
                registers[i] = deviceRegs[i]; // Read registers.
            }
            else
            {
                deviceRegs[i] = registers[i]; // Write registers.into shaddow
                if ((i != PC) && (i != SR) && (i != R3))    // Write register immediately if not PC or SR or R3
                  if (writeReg(i) != STATUS_OK)
                  {
                      RET_ERR(WRITE_REGISTER_ERR); // Could not write Regiser
                  }
            }
        }
    }

    RET_OK;
}


/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Erase(LONG type, LONG address, LONG length);

Description:
 Erase the device FLASH memory.

Parameters:
 type: ERASE_SEGMENT: Erase the segment containing 'address'.
       ERASE_MAIN:    Erase the Main memory.
       ERASE_ALL:     Erase the Main and Information memories.
 address:             Starting address of erase check operation. Must be word aligned.
 length:              Length of erase check operation (even number of bytes).

Returns:
 STATUS_OK:    The device FLASH memory was erased.
 STATUS_ERROR: The device FLASH memory was not erased.

Error codes:
 PARAMETER_ERR
 PRESERVE_RESTORE_ERR
 FREQUENCY_ERR
 ERASE_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. Use MSP430_Read_Memory() to determine the address of a failed erase operation.
 3. If address+length extends beyond the segment containing address, intermediate segments are erased
    (and checked).
 4. Following erasure of the FLASH, the device oscillator will be set to approx. 2.1MHz.
*/
STATUS_T WINAPI MSP430_Erase(LONG type, LONG address, LONG length)
{
    STATUS_T result = STATUS_OK;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Erase type: 0x%4.4X   address: 0x%4.4X   length: 0x%4.4X  \n",timeDiff, type, address, length);
    fclose(tracefile);
#endif


    if ((deviceId == DEVICE_UNKNOWN) || ((type != ERASE_SEGMENT) && (type != ERASE_MAIN) && (type != ERASE_ALL)) ||
        (address < 0) || ((address + length) > (LONG)(deviceInternal.mainEnd + 1)))
    {
        RET_ERR(PARAMETER_ERR);
    }

    result = eraseFlash(type, address, length); 

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    fprintf(tracefile, " -> Exited OK\n");
    fclose(tracefile);
#endif

    return (result);
    
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Memory(LONG address, CHAR* buffer, LONG count, LONG rw);

Description:
 Read and write the device memory. "Device memory" includes the Special Function Registers (i.e., peripheral
 registers), RAM, Information (FLASH) memory, and Main (FLASH) memory.

Parameters:
 address: The starting address of the device memory to be read or written.
 buffer:  The buffer into which device memory is read, or from which device memory is written.
 count:   The number of bytes of device memory read or written.
 rw:      Specify a read (READ) or write (WRITE) operation.

Returns:
 STATUS_OK:    The memory operation encountered no errors.
 STATUS_ERROR: The memory operation encountered errors.

Error codes:
 PARAMETER_ERR
 READ_MEMORY_ERR
 WRITE_MEMORY_ERR
 PRESERVE_RESTORE_ERR
 FREQUENCY_ERR

Notes:
 1. A data write operation can be verified by performing a data read operation followed by a comparison at
    the application level.
 2. Usually, the FLASH memory must be erased before the write to FLASH memory operation is performed.
    *** The write to FLASH memory operation DOES NOT erase the FLASH before the operation is performed. ***
 3. Following a write to FLASH, the device oscillator will be set to approx. 2.1MHz.
*/
STATUS_T WINAPI MSP430_Memory(LONG address, CHAR* buffer, LONG count, LONG rw)
{
  WORD doneCount = 0;
  WORD memValue;
  BOOL adjustLast = FALSE;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Memory  address: 0x%4.4X   count: 0x%4.4X   rw: 0x%4.4X  *buffer: 0x%10.10x \n",timeDiff, address, count, rw, buffer);
    fclose(tracefile);
#endif

  if ((address < 0) || ((address + count) > (LONG)(deviceInternal.mainEnd + 1)) || !buffer)
  {
      RET_ERR(PARAMETER_ERR);
  }

  if (rw)
  { // Read.
      if (address < WORD_REG_START_ADDR) // Process all byte addresses first.
      {
          doneCount = WORD_REG_START_ADDR - (WORD) address;
          if (count < doneCount)
              doneCount = (WORD) count;
          
          if (readByte((WORD) address, doneCount, buffer) != STATUS_OK)
          {
              RET_ERR(READ_MEMORY_ERR);
          }

          // special handling of the I2CDR Register which could be r/w in word and byte mode
          if (deviceHasI2C())
          {
              BYTE temp[10];
              if (readByte((WORD) bI2CCTL, 6, temp) != STATUS_OK)
              {
                  RET_ERR(READ_MEMORY_ERR);
              }
              // test if I2C is enabled and WORD mode is active
              if ((temp[0] & 0x25) && (temp[1] & 0x80))
              if ((address <= bI2CDR) && ((address + count) > bI2CDR)) // Watchdog time
              {
                  if (readWord((WORD) bI2CDR, 1, (unsigned short *) temp) != STATUS_OK)
                  {
                      RET_ERR(READ_MEMORY_ERR);
                  }
                  buffer[bI2CDR - address    ] = temp[0];
                  buffer[bI2CDR - address + 1] = temp[1];
              }
          }

      }
      else if (address & 1) // Otherwise force word alignment if not.
      {
          if (readWord((WORD) (address - 1), 1, &memValue) != STATUS_OK)
          {
              RET_ERR(READ_MEMORY_ERR);
          }
          *buffer = (memValue >> 8) & 0xFF;
          doneCount = 1;
      }
      
      // And now the word memory.   
      if ((count - doneCount) > 0)
      {
          if ((address + count) & 1) // The last address has to be aligned.
          {
              adjustLast = TRUE;
              count--;
          }
          
/*        // read 0x200 sperately
          if ((address + doneCount) == (LAST_PERIPHERAL_ADDR + 1)) // The last address has to be aligned.
          {
           if (readWord((WORD) (address + doneCount), (WORD) 1,
              (WORD*) (buffer + doneCount)) != STATUS_OK)
           {
              RET_ERR(READ_MEMORY_ERR);
           }
              doneCount++;
              doneCount++;
          }
*/
        if (readWord((WORD) (address + doneCount), (WORD) ((count - doneCount) / 2),
              (WORD*) (buffer + doneCount)) != STATUS_OK)
          {
              RET_ERR(READ_MEMORY_ERR);
          }
          
          if (adjustLast)       
          {
              if (readWord((WORD) (address + count), 1, &memValue) != STATUS_OK)
              {
                  RET_ERR(READ_MEMORY_ERR);
              }
              *(buffer + count) = memValue & 0xFF;
          }
      }

      if (address < NSFRSMASKED) // Memory contents from 0x0 to 0x(NSFRSMASKED - 1) require masking.
      {
          int i;
          
          for (i = 0; i < NSFRSMASKED; i++)
          {
              if ((address <= i) && (address + count > i))
              {
                  buffer[i - address] &= deviceSfrMasks(i); // Mask Special Function Registers.
              }
          }
      }
      
      // If any of the special registers are held by the driver, display to the user what the value would be.
      if ((address <= WDT) && ((address + count) > WDT)) // Watchdog time
      {
          buffer[WDT - address] = wWDTCTL & 0xff;
      }
      if (deviceHasDTCBug())
      {
        if ((address <= bADC10DTC0) && (bADC10DTC0 < address + count)) // ADC10DTC0
        {
           buffer[bADC10DTC0 - address]= (BYTE) (savedAdc10dtc0);
        }
        if ((address <= (wADC10CTL0 + 0)) && ((wADC10CTL0 + 0) < (address + count))) // low byte ADC10CTL0
        {
           buffer[wADC10CTL0 + 0 - address] = (BYTE) (savedAdc10ctl0);
        }
        if ((address <= (wADC10CTL0 + 1)) && ((wADC10CTL0 + 1) < (address + count))) // high byte ADC10CTL0
        {
           buffer[wADC10CTL0 + 1 - address] = (BYTE) (savedAdc10ctl0 >> 8);
        }
        if ((address <= (wADC10CTL1 + 0)) && ((wADC10CTL1 + 0) < (address + count))) // low byte ADC10CTL1
        {
           buffer[wADC10CTL1 + 0 - address] = (BYTE) (savedAdc10ctl1);
        }
        if ((address <= (wADC10CTL1 + 1)) && ((wADC10CTL1 + 1) < (address + count))) // high byte ADC10CTL1
        {
           buffer[wADC10CTL1 + 1 - address] = (BYTE) (savedAdc10ctl1 >> 8);
        }
     }
  } // Read.
  else
  { // Write.
      if (((address >= 0                ) && ((address + count - 1) <= deviceRamEnd() )) ||
          ((address >= deviceRam2Start()) && ((address + count - 1) <= deviceRam2End())) )
      { // Write in io/ram.
          int i = 0;
          
          // Check buffered (saved) module registers in order from low to high addresses.
          // The registers referenced below are not written until GO or SINGLE STEP.
          
          // Will ADC10DTC0 be written?
          if ((address <= bADC10DTC0) && (bADC10DTC0 < address + count))
          {
              // write part before ADC10DTC0
              if (writeBlock((WORD) (address + i), buffer + i, (WORD) (bADC10DTC0 - (address + i))) != STATUS_OK)
              {
                  RET_ERR(WRITE_MEMORY_ERR);
              }
              savedAdc10dtc0 = buffer[bADC10DTC0 - address];
              i = bADC10DTC0 - address + 1;
          }

          
          // Will I2CDR be written?
          if ((address <= bI2CDR) && (bI2CDR < address + count))
          {
              // special handling of the I2CDR Register which could be r/w in word and byte mode
              if (deviceHasI2C())
              {
                  BYTE temp[10];
                  if (readByte((WORD) bI2CCTL, 6, temp) != STATUS_OK)
                  {
                      RET_ERR(READ_MEMORY_ERR);
                  }
                  // test if I2C is enabled and WORD mode is active
                  if ((temp[0] & 0x25) && (temp[1] & 0x80))
                  if ((address <= bI2CDR) && ((address + count) > bI2CDR)) // Watchdog time
                  {
                      // write part before I2CDR
                      if (writeBlock((WORD) (address + i), buffer + i, (WORD) (bI2CDR - (address + i))) != STATUS_OK)
                      {
                          RET_ERR(WRITE_MEMORY_ERR);
                      }
                      i = bI2CDR - address + 2;
                      if (writeWord((WORD) bI2CDR, 1, (WORD*)(buffer + (bI2CDR - address))) != STATUS_OK)
                      {
                          RET_ERR(WRITE_MEMORY_ERR);
                      }
                  }
              }
          }

          // Will WDT be written?
          if ((address + i) <= WDT + 1)
          {
              if ((address + count) <= WDT)
              {
                  if (writeBlock((WORD) (address + i), buffer + i, (WORD) (count - i)) != STATUS_OK)
                  {
                      RET_ERR(WRITE_MEMORY_ERR);
                  }

                  RET_OK;
              }
              else
              {
                  // write part before WDTCTL
                  if (writeBlock((WORD) (address + i), buffer + i, (WORD) (WDT - (address + i))) != STATUS_OK)
                  {
                      RET_ERR(WRITE_MEMORY_ERR);
                  }
                  i = WDT - address;
                  if (i == -1)
                  {
                      // high byte only
                      buffer[WDT - address + 1] = WDTPW; // Insure WDT password is present.
                      i = 1;
                  }
                  else if (i + 2 < count)
                  {  
                      // low and high byte
                      buffer[WDT - address + 1] = WDTPW; // Insure WDT password is present.
                      // Update saved WDT value.
                      wWDTCTL = (WORD)(WDTPW << 8) | (BYTE)buffer[WDT - address]; 
                      i += 2;
                  }
                  else
                  {
                      // low byte only
                      // Update saved WDT value.
                      wWDTCTL = (WORD)(WDTPW << 8) | (BYTE)buffer[WDT - address];
                      
                      RET_OK;
                  }
              }
          }
          
        if (deviceHasDTCBug())
        {
           // Will ADC10CTL0 be written?
           if (handleSavedModuleReg((WORD) address, buffer, (WORD) count, &i, wADC10CTL0, &savedAdc10ctl0) != STATUS_OK)
           {
              RET_ERR(WRITE_MEMORY_ERR);
           }
          
           // Will ADC10CTL1 be written?
           if (handleSavedModuleReg((WORD) address, buffer, (WORD) count, &i, wADC10CTL1, &savedAdc10ctl1) != STATUS_OK)
           {
              RET_ERR(WRITE_MEMORY_ERR);
           }
          
        }

        if (writeBlock((WORD) (address + i), buffer + i, (WORD) (count - i)) != STATUS_OK)
        {
           RET_ERR(WRITE_MEMORY_ERR);
        }

      } // Write in io/ram.
      else
      {
          if (((address >= (LONG)deviceInternal.infoStart) && ((address + count - 1) <= (LONG)deviceInternal.infoEnd)) || // Info Mem
              ((address >= (LONG)deviceInternal.mainStart) && ((address + count - 1) <= (LONG)deviceInternal.mainEnd))  || // Main Mem 
              ((address >= (LONG)deviceInternal.infoStart) && ((address + count - 1) <= (LONG)deviceInternal.mainEnd) && 
                                                       ((deviceInternal.infoEnd + 1) == deviceInternal.mainStart)) ) // Both Mem 
          {   
              if(writeFlash((LONG) address, buffer, (LONG) count) != STATUS_OK)
                RET_ERR(WRITE_MEMORY_ERR);
          }
          else
          {
              RET_ERR(WRITE_MEMORY_ERR); // Write starts and/or ends outside of valid memory.
          }

      }
  } // Write.

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    fprintf(tracefile, " -> Exited OK\n");
    fclose(tracefile);
#endif

  RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Breakpoint(LONG number, LONG address);

Description:
 An Instruction Fetch breakpoint is set on address "address". Breakpoint block number "number" is used.

Parameters:
 number:  The number of the breakpoint block to use (>= 0).
 address: The address of the Instruction Fetch breakpoint.

Returns:
 STATUS_OK:    The breakpoint was set.
 STATUS_ERROR: The breakpoint was not set.

Error codes:
 PARAMETER_ERR
 BREAKPOINT_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. Setting a breakpoint in an invalid breakpoint block will return STATUS_ERROR. This mechanism can be
    employed to determine how many breakpoint blocks a device supports. Or reference device.nBreakpoints.
 3. A breakpoint block can be disabled by setting its breakpoint address to zero (0x0).
 4. It is the responsibility of the application to maintain what breakpoints (if any) are set in what
    breakpoint blocks (as it's not possible to read the breakpoint address of a breakpoint block).
 5. A breakpoint set on odd address N will be set on even address N-1.
 6. A breakpoint is enabled only prior to RUN_TO_BREAKPOINT.
 7. Hitting a breakpoint does not clear/disable it; it must be cleared/disabled explicitly.
*/
STATUS_T WINAPI MSP430_Breakpoint(LONG number, LONG address)
{
    WORD bpEnable;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Breakpoint number: 0x%4.4X   address: 0x%4.4X  \n",timeDiff, number, address);
      fclose(tracefile);
#endif

    if ((deviceId == DEVICE_UNKNOWN) || (number < 0) || (number >= deviceNBreakpoints()) ||
        (address < 0) || (address > (LONG)deviceInternal.mainEnd))
    {
        RET_ERR(PARAMETER_ERR);
    }

    if (deviceEmulation() != EMEX_NONE)
    {
        // Set breakpoint address.
        if ((jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL) ||
             jtagCommand(J_WORD, (WORD) (number * MX_BLKSIZE + MX_BP + MX_WRITE), NULL) ||
             jtagCommand(J_WORD, (WORD) address, NULL) ||
          jtagCommand(J_WORD, (WORD) (number * MX_BLKSIZE + MX_CNTRL + MX_WRITE), NULL)            ||
         // Break on memory address bus during instruction fetch.
         // Special consideration for future data breakpoints and the F12x2:
         // - Must handle DMA access (Bit 6 of control register).
         // - Read/Write Enable (Bit 5 of control register) is not implemented in the F12x2.
          jtagCommand(J_WORD, BPCNTRL_EQ | BPCNTRL_RW_DISABLE | BPCNTRL_IF | BPCNTRL_MAB, NULL) ||
          jtagCommand(J_WORD, (WORD) (number * MX_BLKSIZE + MX_MASK + MX_WRITE), NULL)            ||
          jtagCommand(J_WORD, BPMASK_WORD, NULL)                                       ||
          jtagCommand(J_WORD, (WORD) (number * MX_BLKSIZE + MX_COMB + MX_WRITE), NULL)            ||
          jtagCommand(J_WORD, (WORD) (0x01 << number ), NULL) /* Breakpoint */) != STATUS_OK)

        {
            RET_ERR(BREAKPOINT_ERR);
        }
      if ((jtagCommand(J_WORD, (WORD) (MX_CPUSTOP + MX_READ), NULL) || // Read BP status .
          jtagCommand(J_WORD, 0x0000, &bpEnable)                     ||  
          jtagCommand(J_WORD, (WORD) (MX_CPUSTOP + MX_WRITE), NULL) || // Enable breakpoint(s).
          jtagCommand(J_WORD, (WORD) (bpEnable | (WORD) (0x01 << number)), NULL)   ) != STATUS_OK)
      {
         return (STATUS_ERROR);
      }
    }
    else
    {
        BYTE bpInstr[] = { IR_BP1_16BIT, IR_BP2_16BIT };
        WORD bpCntl = 0x0000; // Enable writing into breakpoint registers.
        
        if ((jtagCommand(J_INSTR, IR_BP_CNTL_16BIT, NULL) || // Clear enable bits to permit writing.
             jtagCommand(J_WORD, bpCntl, NULL)            ||
             jtagCommand(J_INSTR, bpInstr[number], NULL)  || // Write address of breakpoint.
             jtagCommand(J_WORD, (WORD) address, NULL)) != STATUS_OK)
        {
            RET_ERR(BREAKPOINT_ERR);
        }
    }

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Run(LONG mode, LONG releaseJTAG);

Description:
 Run the device using the specified mode. JTAG control signals are optionally released.

Parameters:
 mode:        The specified run mode.
               FREE_RUN:          Run the device. Set breakpoints (if any) are disabled.
               SINGLE_STEP:       A single device instruction is executed. Interrupt processing is supported.
               RUN_TO_BREAKPOINT: Run the device. Set breakpoints (if any) are enabled.
 releaseJTAG: The JTAG control signals are released when TRUE.

Returns:
 STATUS_OK:    The run operation encountered no errors.
 STATUS_ERROR: The run operation encountered errors.

Error codes:
 DEVICE_UNKNOWN_ERR
 STEP_ERR
 RUN_ERR
 BREAKPOINT_ERR
 PARAMETER_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. Use MSP430_State() to determine the device state after an MSP430_Run() operation.
 3. Use MSP430_State() to update the device registers before using MSP430_Registers() to read the
    device registers.
*/
STATUS_T WINAPI MSP430_Run(LONG mode, LONG releaseJTAG)
{
    WORD wInstruction = 0;
    WORD wSR_Stack;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Run mode: 0x%4.4X   releaseJTAG: 0x%4.4X  \n",timeDiff, mode, releaseJTAG);
      fclose(tracefile);
#endif

    if (deviceId == DEVICE_UNKNOWN)
    {
        RET_ERR(DEVICE_UNKNOWN_ERR);
    }

    switch (mode)
    {
    case SINGLE_STEP:
        // The CPU cannot be single stepped if it is OFF (Low Power Mode) and interrupts are disabled.
        // Since, while C-SPY has control of the device, CPUOff is cleared, use the saved state of the SR.
        if ((deviceRegs[SR] & STATUS_REG_CPUOFF) && // Low power mode is active.
            (!(deviceRegs[SR] & STATUS_REG_GIE)))   // Interrupts are disabled.
        {
            RET_ERR(STEP_ERR);
        }

        extraStep = FALSE;

        readWord((WORD) deviceRegs[SP], 1, &wSR_Stack);
        readWord((WORD) deviceRegs[PC], 1, &wInstruction);
        
        if (deviceEmulation() != EMEX_NONE)
        {
            WORD bpCntrlType = BPCNTRL_IF; // Step to the next instruction fetch if the CPU is not OFF.
            
            if (deviceRegs[SR] & STATUS_REG_CPUOFF)
            {
                // If the CPU is OFF, only step after the CPU has been awakened by an interrupt.
                // This permits single step to work when the CPU is in LPM0 and 1 (as well as 2-4).
                bpCntrlType = BPCNTRL_NIF;
                // Emex requires an additional step when the CPU is OFF (but only if there is not a pending interrupt).
                {
                    WORD cntrlSig;
                    
                    if ((jtagCommand(J_INSTR, IR_CNTRL_SIG_CAPTURE, NULL)       ||
                         jtagCommand(J_WORD, 0x0000, &cntrlSig)                   ) != STATUS_OK)
                    {
                        RET_ERR(STEP_ERR); // Could not determine CPU state prior to single step!
                    }
                    
                    if (!(cntrlSig & CNTRL_SIG_INTR_REQ))
                    {
                        extraStep = TRUE;
                    }
                }
            }
            if ((jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL)              ||
             MSP430_EEM_Read_Register_Test((WORD) (MX_CNTRL), &brkPnt0Bckup.cntrl) || // Backup breakpoint block 0
             MSP430_EEM_Read_Register_Test((WORD) (MX_MASK),  &brkPnt0Bckup.mask) ||
             MSP430_EEM_Read_Register_Test((WORD) (MX_COMB),  &brkPnt0Bckup.comb) ||
             MSP430_EEM_Read_Register_Test((WORD) (MX_CPUSTOP),  &brkPnt0Bckup.cpuStop) ||
             MSP430_EEM_Read_Register_Test((WORD) (MX_BP), &brkPnt0Bckup.value) ||

                 jtagCommand(J_WORD, (WORD) (MX_CNTRL + MX_WRITE), NULL)        || // Use breakpoint block 0 to single step.
                 // Break on memory address bus during instruction fetch.
                 // Special consideration for future data breakpoints and the F12x2:
                 // - Must handle DMA access (Bit 6 of control register).
                 // - Read/Write Enable (Bit 5 of control register) is not implemented in the F12x2.
                 jtagCommand(J_WORD, (WORD) (BPCNTRL_EQ | BPCNTRL_RW_DISABLE | bpCntrlType | BPCNTRL_MAB), NULL) ||
                 jtagCommand(J_WORD, (WORD) (MX_MASK + MX_WRITE), NULL)         ||
                 jtagCommand(J_WORD, BPMASK_DONTCARE, NULL)                     || // Use DONTCARE to single step.
                 jtagCommand(J_WORD, (WORD) (MX_COMB + MX_WRITE), NULL)         ||
                 jtagCommand(J_WORD, 0x01, NULL)                                || // Breakpoint 0.
                 jtagCommand(J_WORD, (WORD) (MX_CPUSTOP + MX_WRITE), NULL)      || // Enable breakpoint.
                 jtagCommand(J_WORD, 0x01, NULL)    /* Trigger 0. */              ) != STATUS_OK)
            {
                RET_ERR(STEP_ERR); // Could not perform single step!
            }

         brkPnt0Bckup.valid=1;

/*            if (deviceClockControl())
            {
                if ((jtagCommand(J_WORD, (WORD) (MX_GCLKCTRL + MX_WRITE), NULL) || // Configure clock control.
                     jtagCommand(J_WORD, (WORD) 0, NULL)                   ) != STATUS_OK)
                {
                    RET_ERR(STEP_ERR); // Could not perform single step!
                }
            }
*/
//-----------------------------------------------------------------------------------------
         // copied this part from 'setBreakpointsActive' to fix running clock after single step
            if (deviceClockControl() != GCC_NONE)
            {
               if ((jtagCommand(J_WORD, (WORD) (MX_GCLKCTRL + MX_WRITE), NULL) || // Configure clock control.
                   jtagCommand(J_WORD, (WORD) clkCntrl, NULL)                 ) != STATUS_OK)
               {
                  return (STATUS_ERROR);
               }
            }
//-----------------------------------------------------------------------------------------

            if (writeAllDeviceRegisters(deviceRegs) != STATUS_OK) // Restore all device registers.
            {
                RET_ERR(STEP_ERR);
            }
            if ((jtagCommand(J_INSTR, IR_EMEX_WRITE_CONTROL, NULL)              || // Single step (release the CPU).
                 jtagCommand(J_WORD, (WORD) (genCntrl | CLEAR_STOP | EEM_EN), NULL)) != STATUS_OK)
            {
                RET_ERR(STEP_ERR); // Could not perform single step!
            }
        } // if (deviceEmulation != EMEX_NONE)
        else // (Older) Device does not have EMEX.
        {
            if (writeAllDeviceRegisters(deviceRegs) != STATUS_OK) // Restore all device registers.
            {
                RET_ERR(STEP_ERR);
            }
            if ((jtagCommand(J_INSTR, IR_BP_CNTL_16BIT, NULL)                   || // Single step (release the CPU).
                 jtagCommand(J_WORD, 0x0080, NULL)                                ) != STATUS_OK)
            {
                RET_ERR(STEP_ERR); // Could not perform single step!
            }
        } // (Older) Device does not have EMEX.
        
        if (releaseCPU() != STATUS_OK) // Perform the single step.
        {
            RET_ERR(STEP_ERR);
        }
        // Fake the system into thinking that hardware breakpoints are used for single step.
        // This cause procState to be updated in syncCPUAfterDebug() as expected.
        hardwareBPsUsed = TRUE;

        if (CPUCycles == -1) { CPUCycles = 0; }
        if (CPUCycles < -1) { CPUCycles += (2147483647); CPUCycles++; } // Handle overflow.
        
        {
            WORD cntrlSig;
            if ((jtagCommand(J_INSTR, IR_CNTRL_SIG_CAPTURE, NULL)       ||
                 jtagCommand(J_WORD, 0x0000, &cntrlSig)) != STATUS_OK)
            {
                   RET_ERR(STEP_ERR); // Could not determine CPU state prior to single step!
            }
            if ((cntrlSig & CNTRL_SIG_INTR_REQ) && (deviceRegs[SR] & STATUS_REG_GIE)) // Interrupt pending and enabled.
            {
                CPUCycles += 6; // Interrupt is handled.
            }
            else
            {   
                if (!(deviceRegs[SR] & STATUS_REG_CPUOFF)) // Lot LPM.
                {
                    if (wInstruction == 0x1300) // RETI
                    {
                        {
                            if (cntrlSig & CNTRL_SIG_INTR_REQ)
                            {
                                if ((wSR_Stack & STATUS_REG_GIE)) { CPUCycles += 6; } // RETI + next intr. request.
    //                            if ((wSR_Stack & STATUS_REG_CPUOFF)) { CPUCycles += 6; } // RETI + next intr. request.
                            }
                        }

                    }
                    GetCycles(wInstruction);
                }
            }
        }
        break; // case SINGLE_STEP:
    case FREE_RUN:
    case RUN_TO_BREAKPOINT:
        CPUCycles = -1; // Reset count of CPU cycles executed.
        if (writeAllDeviceRegisters(deviceRegs) != STATUS_OK) // Restore all device registers.
        {
            RET_ERR(RUN_ERR);
        }
        if (mode == RUN_TO_BREAKPOINT)
        {
            if (setBreakpointsActive(deviceNBreakpoints(), &bpState) != STATUS_OK) // Activate all breakpoints.
            {
                RET_ERR(BREAKPOINT_ERR);
            }
        }
        if (releaseCPU() != STATUS_OK) // Start device execution (and retain the JTAG control signals).
        {
            RET_ERR(RUN_ERR);
        }
        if (releaseJTAG)
        {
            if (HIL_Release() != STATUS_OK) // Release the JTAG control signals.
            {
                RET_ERR(RUN_ERR);
            }
            jtagReleased = TRUE;
        }
        else
        {
            jtagReleased = FALSE;
        }
        break;
    default:
        RET_ERR(PARAMETER_ERR);
        //break;
    }
    runMode = mode;

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_State(LONG* state, LONG stop, LONG* pCPUCycles);

Description:
 Determine the state of the device following an MSP430_Run() operation. The device is optionally stopped.
 During single step operations, a count of CPU cycles executed is maintained.

Parameters:
 state:      The device state:
              STOPPED:              The device is stopped.
              RUNNING:              The device is running or is being single stepped.
              SINGLE_STEP_COMPLETE: The device is stopped after completing the single step operation.
              BREAKPOINT_HIT:       The device is stopped as a result of hitting an enabled breakpoint.
 stop:       The device is stopped when TRUE.
 pCPUCycles: The cumulative number of CPU cycles executed during single step (or -1 if invalid).

Returns:
 STATUS_OK:    The device state was determined.
 STATUS_ERROR: The device state was not determined.

Error codes:
 PARAMETER_ERR
 STATE_ERR
 STEP_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. To determine the device state, the JTAG control signals are required. If the JTAG control signals were
    released, the signals are reconnected. If the state is RUNNING and the JTAG control signals had been
    released and stop is FALSE, the JTAG control signals are released after determining the state.
 3. The number of CPU cycles executed is zeroed when the device is reset (MSP430_Reset()), and when the
    device is run (MSP430_Run()) with FREE_RUN and RUN_TO_BREAKPOINT.
 4. Use MSP430_State() to update the device registers before using MSP430_Registers() to read the
    device registers.
*/
STATUS_T WINAPI MSP430_State(LONG* state, LONG stop, LONG* pCPUCycles)
{
    WORD procState;
    WORD pc;
    BOOL clearedCpuOff;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_State state: 0x%4.4X   stop: 0x%4.4X   pCPUCycles: 0x%4.4X  \n",timeDiff, *state, stop, *pCPUCycles);
      fclose(tracefile);
#endif


#if defined(_MSC_VER)   // compiled with Microsoft
    {   // this sequence is used to avoid to high access to the parallel port
        static DWORD lastCheckTime = 0;
        static DWORD CheckTimeDiff = 0;
        if ((!stop) && (*state == RUNNING))
        {
            CheckTimeDiff = timeGetTime() - lastCheckTime;
            if (CheckTimeDiff <= 100) 
            {    RET_OK;} // Do nothing if check is to short 
            else
            {    lastCheckTime = timeGetTime();}
        }
    }
#endif


    if ((deviceId == DEVICE_UNKNOWN) || !state)
    {
        RET_ERR(PARAMETER_ERR);
    }

    if (pCPUCycles != NULL) { *pCPUCycles = CPUCycles; } // Return count of CPU cycles executed.

    // If the device is already STOPPED, indicate this. This logic permits MSP430_State() to be called repeatedly.
    if (runMode == STOPPED)
    {
        *state = STOPPED;

        RET_OK;
    }

    *state = RUNNING;

    if (jtagReleased) // Enable JTAG connections if released.
    {
        if (HIL_Connect() != STATUS_OK)
        {
            RET_ERR(STATE_ERR);
        }
        HIL_DelayMSec(20);       // wait 20 ms
        testConnect_TAPReset();  // Reset TAP
    }

    // If breakpoints set or single stepping.
    if (((runMode == RUN_TO_BREAKPOINT) && hardwareBPsUsed) || (runMode == SINGLE_STEP))
    {
        if (deviceEmulation() != EMEX_NONE)
        {
            if ((jtagCommand(J_INSTR, IR_EMEX_READ_CONTROL, NULL) || // Poll for breakpoint/ss hit.
                 jtagCommand(J_WORD, 0x0000, &procState)            ) != STATUS_OK)
            {
                RET_ERR(STATE_ERR) ; // Could not poll for breakpoint/ss hit!
            }
            if (procState & EEM_STOPPED) // Breakpoint/ss hit!
            {
                if (extraStep)
                {
                    if ((deviceRegs[SR] & STATUS_REG_GIE)) // Interrupt pending and enabled.
                    {
                        CPUCycles += 6; // Interrupt is handled.
                    }
                    extraStep = FALSE;
                    if (releaseCPU() != STATUS_OK)
                    {
                        RET_ERR(STEP_ERR); // Could not perform extra single step!
                    }
                    // The state is RUNNING.
                }
                else
                {
                    *state = (runMode == RUN_TO_BREAKPOINT) ? BREAKPOINT_HIT : SINGLE_STEP_COMPLETE;
                    stop = TRUE; // Force the device to stop.
                }
            }
        }
        else
        {
            if ((jtagCommand(J_INSTR, IR_BP_CNTL_CAPTURE, NULL)   || // Poll for breakpoint/ss hit.
                 jtagCommand(J_WORD, 0x0000, &procState)            ) != STATUS_OK)
            {
                RET_ERR(STATE_ERR) ; // Could not poll for breakpoint/ss hit!
            }
            if (runMode == RUN_TO_BREAKPOINT)
            {
                if ((procState & 0x0003) != (bpState & 0x0003))
                {
                    *state = BREAKPOINT_HIT; // Breakpoint hit!
                    stop = TRUE; // Force the device to stop.
                }
            }
            else
            {
                if (!(procState & 0x0080))
                {
                    *state = SINGLE_STEP_COMPLETE; // Single step complete!
                    stop = TRUE; // Force the device to stop.
                }
            }
            if (stop)
            {
                if ((jtagCommand(J_INSTR, IR_BP_CNTL_16BIT, NULL) || // Reset breakpoint(s) (if any) or single step.
                     jtagCommand(J_WORD, 0x0000, NULL)              ) != STATUS_OK)
                {
                    RET_ERR(STATE_ERR); // Could not recover from breakpoint(s) or single step!
                }
            }
        }
    } // If breakpoints set or single stepping.

    if (stop) // Force stop and synchronization.
    {
        // Synchronize with the CPU. Read/Write will be set. Disable Emex if present.
      if ((runMode == SINGLE_STEP)&&brkPnt0Bckup.valid) //Restore Breakpoint 0 only if Backup is valid and single step completed
      {
         if ((jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL)              ||
             MSP430_EEM_Write_Register((WORD) (MX_CNTRL), brkPnt0Bckup.cntrl) || // Backup breakpoint block 0
             MSP430_EEM_Write_Register((WORD) (MX_MASK),  brkPnt0Bckup.mask) ||
             MSP430_EEM_Write_Register((WORD) (MX_COMB),  brkPnt0Bckup.comb) ||
             MSP430_EEM_Write_Register((WORD) (MX_CPUSTOP),  brkPnt0Bckup.cpuStop) ||
             MSP430_EEM_Write_Register((WORD) (MX_BP), brkPnt0Bckup.value)) != STATUS_OK)
         {
            RET_ERR(STEP_ERR);
         }
         brkPnt0Bckup.valid = 0;
      }
        if (syncCPUAfterDebug(&procState, &pc, &clearedCpuOff) != STATUS_OK)
        {
            RET_ERR(STATE_ERR);
        }
        if (readAllDeviceRegisters(deviceRegs, FALSE) != STATUS_OK) // Read the device registers.
        {
            RET_ERR(STATE_ERR);
        }
        deviceRegs[PC] = pc; // And adjust the PC and SR registers.
        if (clearedCpuOff)
        {
            deviceRegs[SR] |= STATUS_REG_CPUOFF;
        }

        // If the device was RUNNING, flag it as STOPPED. Otherwise, leave it flagged as BREAKPOINT_HIT or
        // SINGLE_STEP_COMPLETE.
        if (*state == RUNNING)
        {
            *state = STOPPED;
        }
        runMode = STOPPED;
        hardwareBPsUsed = FALSE; // Reset hardware breakpoints used.
        jtagReleased = FALSE; // Do not release JTAG (if it was released).
    }
    
    if (jtagReleased)
    {
        if (HIL_Release() != STATUS_OK) // Release the JTAG connections if they were originally released.
        {
            RET_ERR(STATE_ERR);
        }
    }

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_Secure(void);

Description:
 The device is secured (i.e., the JTAG security fuse is blown).

 *** The sample HIL.dll that interfaces to the MSP430 via the FET Interface Module connected
 *** to the parallel port of the PC >>> DOES NOT <<< implement the functionality required to
 *** blow the device security fuse.

Parameters:

Returns:
 STATUS_OK:    The device was secured.
 STATUS_ERROR: The device was not secured.

Error codes:
 DEVICE_UNKNOWN_ERR
 BLOW_FUSE_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. Once a device is secure, further communications via JTAG is not possible.
 3. This function requires that HIL_TEST_VPP(), HIL_VCC(), and HIL_VPP() be implemented,
    and HIL_JTAG_IR() must support shifting instructions in via TDO.
 4. After MSP430_Secure(), the device is reset with Vcc set to ((MaxOp + MinOp) / 2) volts (*)
    and Vpp set to 0 volts (*). (*) Implementation dependent.
*/
STATUS_T WINAPI MSP430_Secure(void)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Secure \n",timeDiff);
      fclose(tracefile);
#endif

    if (deviceId == DEVICE_UNKNOWN)
    {
        RET_ERR(DEVICE_UNKNOWN_ERR);
    }

    if (deviceVccMinSecure() == 0) // VccMinSecure == 0 implies device has no fuse.
    {
        RET_ERR(BLOW_FUSE_ERR);
    }

    if(!deviceHasTestVpp()) // Device inputs Vpp via TDI rather than via TEST.
    {
        // Switch TDO functionality to TDI.
        cntrlSigHighByte(CNTRL_SIG_TCE1 | CNTRL_SIG_CPU | CNTRL_SIG_SWITCH); 
        HIL_TEST_VPP(FALSE);
    }
    
    if (jtagCommand(J_INSTR, IR_PREPARE_BLOW, NULL) != STATUS_OK) // Prepare to blow the fuse.
    {
        RET_ERR(BLOW_FUSE_ERR);
    }

    // Set Vcc and Vpp.
    HIL_VCC((LONG) (deviceVccMinSecure() + ((deviceVccMaxOp() - deviceVccMinSecure()) / 2)));
    HIL_VPP((LONG) (deviceVppMinSecure() + ((deviceVppMaxSecure() - deviceVppMinSecure()) / 2)));
    if (jtagCommand(J_INSTR, IR_EX_BLOW, NULL) != STATUS_OK) // Blow the fuse.
    {
        RET_ERR(BLOW_FUSE_ERR);
    }
    HIL_DelayMSec(1);
    HIL_VPP(0); // Zero Vpp.
    HIL_VCC((deviceVccMaxOp() + deviceVccMinOp()) / 2); // Set Vcc.
    
    HIL_TEST_VPP(TRUE); // Restore the normal function of TDO and TDI (within the interface).
    // Note: MSP430_Reset() should now return STATUS_ERROR since the JTAG fuse is blown and
    // normal JTAG communications with the device should not be possible.
    MSP430_Reset(ALL_RESETS, FALSE, FALSE); // Reset the device (to restore normal TDO/TDI functionality).
    testConnect(); // Reset the TAP controller.

    { // Test if the fuse has been blown.
        int i;
        WORD cntrlSig;
        
        for (i = 0; i < 3; i++) // First test could be negative.
        {
            if ((jtagCommand(J_INSTR, IR_CNTRL_SIG_CAPTURE, NULL) ||
                 jtagCommand(J_WORD, 0xaaaa, &cntrlSig)             ) != STATUS_OK)
            {
                RET_ERR(BLOW_FUSE_ERR);
            }
            // If the fuse is blown, the device will be in JTAG bypass mode, and data
            // input to the device will be echoed with a one-bit delay.
            if (cntrlSig == 0x5555)
            {
                RET_OK; // Fuse is blown.
            }
        }
    }
    
    RET_ERR(BLOW_FUSE_ERR); // Fuse is not blown.    
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_VerifyMem(LONG StartAddr, LONG Length, CHAR *DataArray);

Description:
 Compare the MSP430 memory and the specified data.

Parameter:
 StartAddr:    Start address of memory to be compared (must be even).
 Length:       Number of bytes to be compared (must be even).
 *DataArray:   Pointer to data array.

Returns:
 STATUS_OK:    The device memory and data compare.
 STATUS_ERROR: The device memory and data do not compare.

Error codes:
 PARAMETER_ERR
 VERIFY_ERR

Notes:
 1. This function does not read the specified memory region and compare it to the data.
    Instead, this function computes a checksum for the specified memory region, and
    then computes a checksum for the data, and finally compares the two checksums.
 2. The device is reset following this operation.
*/
STATUS_T WINAPI MSP430_VerifyMem(LONG StartAddr, LONG Length, CHAR *DataArray)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_VerifyMem StartAddr: 0x%4.4X   Length: 0x%4.4X  \n",timeDiff, StartAddr, Length);
      fclose(tracefile);
#endif

    return (VerifyPSA(StartAddr, Length, &DataArray[0]));
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_EraseCheck(LONG StartAddr, LONG Length);

Description:
 Verify that the specified memory range is erased.

Parameter:
 StartAddr:    Start address of memory to be verified (must be even).
 Length:       Number of BYTEs to be verified (must be even).

Returns:
 STATUS_OK:    The device memory in the specified range is erased.
 STATUS_ERROR: The device memory in the specified range is not erased.

Error codes:
 PARAMETER_ERR
 VERIFY_ERR

Notes:
 1. This function does not read the specified memory region and compare it to erased data.
    Instead, this function computes a checksum for the specified memory region, and
    then computes a checksum for erased data, and finally compares the two checksums.
 2. The device is reset following this operation.
*/
STATUS_T WINAPI MSP430_EraseCheck(LONG StartAddr, LONG Length)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_EraseCheck StartAddr: 0x%4.4X   Length: 0x%4.4X  \n",timeDiff, StartAddr, Length);
      fclose(tracefile);
#endif

    return (VerifyPSA(StartAddr, Length, 0));
}

// Enhanced Emulation Module functions ----------------------------------------

/* ----------------------------------------------------------------------------

   The following functions implement a very low-level interface to the Enhanced Emulation Module (EEM).
   The EEM provides advanced debugging facilities (including advanced breakpoints, trace buffers, etc.).
   Please contact Texas Instruments to obtain the detailed information and device specifications required
   to make use of this mechanism.

   ---------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_EEM_Open(void);

Description:
 Open the Enhanced Emulation Module (EEM).

Parameters:

Returns:
 STATUS_OK:    The EEM was opened.
 STATUS_ERROR: The EEM was not opened.

Error codes:
 DEVICE_UNKNOWN_ERR
 EEM_OPEN_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
*/
STATUS_T WINAPI MSP430_EEM_Open(void)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_EEM_Open \n",timeDiff);
      fclose(tracefile);
#endif

    if ((deviceId == DEVICE_UNKNOWN) || (deviceEmulation() == EMEX_NONE))
    {
        RET_ERR(DEVICE_UNKNOWN_ERR);
    }

    if (jtagCommand(J_INSTR, IR_EMEX_DATA_EXCHANGE, NULL) != STATUS_OK)
    {
        RET_ERR(EEM_OPEN_ERR);
    }
    eemOpen = TRUE;
    
    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 LONG WINAPI MSP430_EEM_Read_Register(LONG address);

Description:
 Read the EEM register at address "address".

Parameters:
 address: The address of the EEM register read.

Returns:
 The value read from the EEM register.

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. No status is returned from this function; use MSP430_EEM_Read_Register_Test() if a status is desired.
 3. MSP430_EEM_Open() will be called if required.
*/
LONG WINAPI MSP430_EEM_Read_Register(LONG address)
{
    WORD read;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_EEM_Read_Register address: 0x%4.4X   \n",timeDiff, address);
      fclose(tracefile);
#endif


    if ((deviceId == DEVICE_UNKNOWN) || (deviceEmulation() == EMEX_NONE) ||
        (address < 0) || (address > (LONG)deviceInternal.mainEnd))
    {
        return (0); // Return 0 if the EEM register cannot be read (since no error can be indicated).
    }

    if (!eemOpen) // Open the EEM if required.
    {
        if (MSP430_EEM_Open() != STATUS_OK)
        {
            return (0);
        }
    }

    if ((jtagCommand(J_WORD, (WORD) (address | 1), NULL)   || // Force to read address (odd).
         jtagCommand(J_WORD, 0x0000, &read)                  ) != STATUS_OK)
    {
        return (0); // Return 0 if the EEM register cannot be read (since no error can be indicated).
    }
    
    return ((LONG) read);
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_EEM_Read_Register_Test(LONG address, LONG* value);

Description:
 Read the EEM register at address "address", and store the read value at "value".

Parameters:
 address: The address of the EEM register read.
 value:   The address of where the read value is stored.

Returns:
 STATUS_OK:    The EEM read register operation encountered no errors.
 STATUS_ERROR: The EEM read register operation encountered errors.

Error codes:
 PARAMETER_ERR
 EEM_OPEN_ERR
 EEM_READ_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. MSP430_EEM_Open() will be called if required.
*/
STATUS_T WINAPI MSP430_EEM_Read_Register_Test(LONG address, LONG* value)
{
    WORD read;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_EEM_Read_Register_Test address: 0x%4.4X   value: 0x%4.4X  \n",timeDiff, address, *value);
      fclose(tracefile);
#endif

    if ((deviceId == DEVICE_UNKNOWN) || (deviceEmulation() == EMEX_NONE) ||
        (address < 0) || (address > (LONG)deviceInternal.mainEnd) || !value)
    {
        RET_ERR(PARAMETER_ERR);
    }

    if (!eemOpen) // Open the EEM if required.
    {
        if (MSP430_EEM_Open() != STATUS_OK)
        {
            RET_ERR(EEM_OPEN_ERR);
        }
    }

    if ((jtagCommand(J_WORD, (WORD) (address | 1), NULL)   || // Force to read address (odd).
         jtagCommand(J_WORD, 0x0000, &read)                  ) != STATUS_OK)
    {
        RET_ERR(EEM_READ_ERR);
    }

    *value = (LONG) read;

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_EEM_Write_Register(LONG address, LONG value);

Description:
 Write the EEM register at address "address" with value "value".

Parameters:
 address: The address of the EEM register written.
 value:   The value to be written into the EEM register.

Returns:
 STATUS_OK:    The EEM write register operation encountered no errors.
 STATUS_ERROR: The EEM write register operation encountered errors.

Error codes:
 PARAMETER_ERR
 EEM_OPEN_ERR
 EEM_WRITE_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
 2. MSP430_EEM_Open() will be called if required.
*/
STATUS_T WINAPI MSP430_EEM_Write_Register(LONG address, LONG value)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_EEM_Write_Register address: 0x%4.4X   value: 0x%4.4X  \n",timeDiff, address, value);
      fclose(tracefile);
#endif

    if ((deviceId == DEVICE_UNKNOWN) || (deviceEmulation() == EMEX_NONE) ||
        (address < 0) || (address > (LONG)deviceInternal.mainEnd))
    {
        RET_ERR(PARAMETER_ERR);
    }

    if (!eemOpen) // Open the EEM if required.
    {
        if (MSP430_EEM_Open() != STATUS_OK)
        {
            RET_ERR(EEM_OPEN_ERR);
        }
    }

    if ((jtagCommand(J_WORD, (WORD) (address & ~1), NULL)  || // Force to write address (even).
         jtagCommand(J_WORD, (WORD) value, NULL)             ) != STATUS_OK)
    {
        RET_ERR(EEM_WRITE_ERR);
    }

    // Filter clock settings and save values 
    if (address == (MX_GCLKCTRL  + MX_WRITE)) 
        clkCntrl   = (WORD) value;
    if (address == (MX_MCLKCNTL0 + MX_WRITE)) 
        mclkCntrl0 = (WORD) value;
//    if (address == (MX_GENCNTRL + MX_WRITE)) 
//        genCntrl = (WORD) value;

    RET_OK;
}

/* ----------------------------------------------------------------------------
Function:
 STATUS_T WINAPI MSP430_EEM_Close(void);

Description:
 Close the Enhanced Emulation Module (EEM).

Parameters:

Returns:
 STATUS_OK:    The EEM was closed.
 STATUS_ERROR: The EEM was not closed.

Error codes:
 DEVICE_UNKNOWN_ERR
 EEM_CLOSE_ERR

Notes:
 1. MSP430_Identify() must have been called prior to calling this function.
*/
STATUS_T WINAPI MSP430_EEM_Close(void)
{
    WORD cntrlSig;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_EEM_Close \n",timeDiff);
      fclose(tracefile);
#endif

    if ((deviceId == DEVICE_UNKNOWN) || (deviceEmulation() == EMEX_NONE))
    {
        RET_ERR(DEVICE_UNKNOWN_ERR);
    }

    // Any JTAG instruction (and not data) will exit the EEM state.
    // IR_BYPASS is not used as this releases the CPU to execute.
    if ((jtagCommand(J_INSTR, IR_CNTRL_SIG_CAPTURE, NULL) ||
         jtagCommand(J_WORD, 0x0000, &cntrlSig)             ) != STATUS_OK)
    {
        RET_ERR(EEM_CLOSE_ERR);
    }
    eemOpen = FALSE;
    
    RET_OK;
}

// Error handling functions ---------------------------------------------------

/* ----------------------------------------------------------------------------
Function:
 LONG WINAPI MSP430_Error_Number(void);

Description:
 Determine the number of the error when an MSP430_xxx() function returns STATUS_ERROR.

Parameter:

Returns:
 The number of the last error.

Notes;
 1. The error number is reset (to NO_ERR) after the error number is returned.
*/
LONG WINAPI MSP430_Error_Number(void)
{
    int tempErrorNumber = errorNumber;

#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Error_Number \n",timeDiff);
      fclose(tracefile);
#endif

    errorNumber = NO_ERR;

    return (tempErrorNumber);
}

/* ----------------------------------------------------------------------------
Function:
 const CHAR* WINAPI MSP430_Error_String(LONG errorNumber);

Description:
 Determine the string associated with errorNumber.

Parameter:
 errorNumber: Error number.

Returns:
 The string associated with errorNumber.

Notes;
*/
#ifndef uController
const CHAR* WINAPI MSP430_Error_String(LONG errorNumber)
{
#ifdef DEBUG_TRACE
    tracefile=fopen(tracefilename,"a");
    timeDiff = timeGetTime() - lastTime;
    lastTime = timeGetTime();
    fprintf(tracefile, "%5d msec MSP430_Error_String errorNumber: 0x%4.4X  \n",timeDiff, errorNumber);
      fclose(tracefile);
#endif

    if ((errorNumber < 0) || (errorNumber >= INVALID_ERR))
    {
        errorNumber = INVALID_ERR;
    }

    return (errorStrings[errorNumber]);
}
#endif
