/*========================================================================*\
|                                                                          |
| Memory.c                                                                 |
|                                                                          |
| This file contains the Memory read/write functions.                      |
|--------------------------------------------------------------------------|
| Project:              MSP430 JTAG interfcae                              |
| Developed using:      MS Visual C++ 5.0                                  |
|--------------------------------------------------------------------------|
| Author:               Dale Wellborn                                      |
| Version:              1.1.3                                              |
| Initial Version:      20 / 07 / 02                                       |
| Last Change:                                                             |
|--------------------------------------------------------------------------|
| Version history:                                                         |
| Version: 1.1.3                                                           |
| 25.08.2003 UPSF handleSavedModuleReg fixed problem with write of         |
|                 ADC10CTLx (high Byte was nor written                     |
| Version: 1.1.5                                                           |
| 01/21/2004 UPSF added F1611 support                                      |
|                                                                          |
|--------------------------------------------------------------------------|
| Designed 2002 by Texas Instruments                                       |
\*========================================================================*/

/*------------------------------------------------------------------------*\
| Remarks:                                                                 |
|                                                                          |
\*------------------------------------------------------------------------*/


// #includes. -----------------------------------------------------------------

#include "MSP/Basic_Types.h"
#include "MSP/HIL.h"
#include "MSP/MSP430_Debug.h"

#include "MSP/Globals.h"
#include "MSP/JTAG.h"
#include "MSP/JTAG_defs.h"
#include "MSP/Memory.h"

// Functions. -----------------------------------------------------------------

STATUS_T handleSavedModuleReg(WORD address, BYTE* buffer, WORD count, int* j, WORD reg_addr, WORD* saved)
{
	int i = *j;

    if ((address + i) <= (reg_addr + 1))
    {
        if ((address + count) <= reg_addr)
        {
            if (writeBlock((WORD) (address + i), (BYTE*) (buffer + i), (WORD) (count - i)) != STATUS_OK)
            {
                return (STATUS_ERROR);
            }
            *j = count;
        }
        else
        {
            if (writeBlock((WORD) (address + i), (BYTE*) (buffer + i), (WORD) (reg_addr - (address + i))) != STATUS_OK)
            {
                return (STATUS_ERROR);
            }
            i = reg_addr - address;
            if (i == -1)
            {
                // high byte only
                // Update saved value.
                *saved = (WORD)(*saved & 0x00ff) | (WORD)(buffer[reg_addr - address + 1] << 8);
                *j = 1;
            }
            else if (i + 2 <= count)
            {  
                // low and high byte
                // Update saved value.
                *saved = (WORD)(buffer[reg_addr - address + 1] << 8) | (BYTE)buffer[reg_addr - address]; 
                *j = i + 2;
            }
            else
            {
                // low byte only
                // Update saved WDT value.
                *saved = (WORD)(*saved & 0xff00) | (BYTE)buffer[reg_addr - address];
                *j = i + 1;
            }
        }
    }
    else
    {
        *j = i;
    }

	return (STATUS_OK);
}

STATUS_T writeBlock(WORD address, BYTE* buffer, WORD count)
{
	WORD doneCount = 0;
	WORD memValue;
	BOOL adjustLast = FALSE;

	if (address < WORD_REG_START_ADDR) // Process all byte addresses first.
	{
		doneCount = WORD_REG_START_ADDR - (WORD) address;
		if (count < doneCount)
			doneCount = (WORD) count;
		if (writeByte((WORD) address, doneCount, buffer) != STATUS_OK)
			return (STATUS_ERROR);
	}
	else if (address & 1) // Otherwise force word alignment if not.
	{
		if (readWord((WORD) (address - 1), 1, &memValue) != STATUS_OK)
			return (STATUS_ERROR);
		memValue = (memValue & 0xFF) + (*buffer << 8);
		if (writeWord((WORD) (address - 1), 1, &memValue) != STATUS_OK)
			return (STATUS_ERROR);
		doneCount = 1;
	}
		  
	// And now the word memory.	
	if ((count - doneCount) > 0)
	{
		if ((address + count) & 1) // The last address has to be aligned.
		{
			adjustLast = TRUE;
			count--;
		}
		if (writeWord((WORD) (address + doneCount), (WORD) ((count - doneCount) / 2),
			(WORD*) (buffer + doneCount)) != STATUS_OK)
			return (STATUS_ERROR);
		if (adjustLast)		
		{
			if (readWord((WORD) (address + count), 1, &memValue) != STATUS_OK)
				return (STATUS_ERROR);
			memValue = (memValue & 0xFF00) + *(buffer + count);
			if (writeWord((WORD) (address + count), 1, &memValue) != STATUS_OK)
				return (STATUS_ERROR);
		}
	}

	return (STATUS_OK);
} // writeBlock()

//-----------------------------------------------------------------------------
// Assumptions:
// 1) The processor is on an instruction load boundary. This implies that TCLK and R/W are high.
// 2) The program counter does not point into RAM.

STATUS_T readWord(WORD Address, WORD Length, WORD* Buffer)
{
	int i;

	if (Length == 0)
	{
		return (STATUS_OK);
	}
	if (Address & 1) // Address must be even.
	{
		return (STATUS_ERROR);
	}
	if (isInstrLoad() != STATUS_OK) // (As an internal check) Must start on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}

	if (Address > LAST_PERIPHERAL_ADDR)
	{
		// Read memory (ram, information, program).
		
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
		HIL_TCLK(0);
		//HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16((WORD) (Address - 2));
		HIL_TCLK(NEG_EDGE);
		HIL_TCLK(NEG_EDGE);
		HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
		if (HIL_JTAG_DR16(0) != Address - 2)
		{
			return (STATUS_ERROR);
		}
		// The next rising TCLK will clock the CPU onto an instruction load boundary.
		
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(JMP_D); // Load "jmp $" instruction for use while HALT'ed.
		HIL_TCLK(NEG_EDGE);
		if (isInstrLoad() != STATUS_OK) // (As an internal check) Must be on an instruction load boundary.
		{
			return (STATUS_ERROR);
		}
		// Enable read from memory and halt the CPU (with the next TCLK).
		// The F12x2 requires that TCLK be low when HALT_JTAG is set.
		cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG);
		HIL_JTAG_IR_(IR_DATA_QUICK);
		
		for (i = 0; i < Length; i++)
		{
			HIL_TCLK(NEG_EDGE);
			Buffer[i] = (WORD) HIL_JTAG_DR16(0); // Read data from memory.
		}
	}
	else
	{
        // Read peripherals. Don't use DATA_QUICK.
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(JMP_D); // jmp $ for use while HALT'ed.
		HIL_TCLK(0);
		// The F12x2 requires that TCLK be low when HALT_JTAG is set.
		cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG); // Halt the CPU.
		HIL_TCLK(NEG_EDGE); // Stop CPU with this clock.

		for (i = 0; i < Length; i++)
		{
			HIL_JTAG_IR_(IR_ADDR_16BIT);
			HIL_JTAG_DR16((WORD) (Address + i * 2));
			HIL_JTAG_IR_(IR_DATA_TO_ADDR);
			HIL_TCLK(NEG_EDGE);
			Buffer[i] = (WORD) HIL_JTAG_DR16(0);
		}
	}

	cntrlSigLowByte(CNTRL_SIG_READ); // Release the CPU.
	// The next rising TCLK will clock the CPU onto an instruction load boundary.

	if (deviceHasJTAGBug())
	{
		WORD pc;

		HIL_JTAG_IR_(IR_ADDR_CAPTURE);
		pc = (WORD) HIL_JTAG_DR16(0);
		// Does the PC point into RAM? Note: Some devices do not completely decode the RAM address space.
		// Consequently, check if the PC is anywhere in the RAM space (and not just where the device has actual RAM).
		if ((pc >= deviceRamStart()) && (pc < deviceRamEnd()))
		{
			// Prevent RAM corruption by setting the PC to an address outside of RAM.
			HIL_JTAG_IR_(IR_DATA_16BIT);
			HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
			HIL_TCLK(NEG_EDGE);
			//HIL_JTAG_IR_(IR_DATA_16BIT);
			HIL_JTAG_DR16(ROM_ADDR);
			HIL_TCLK(NEG_EDGE);
			HIL_TCLK(NEG_EDGE);
			HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
			if (HIL_JTAG_DR16(0) != ROM_ADDR)
			{
				return (STATUS_ERROR);
			}
		}
		// The next rising TCLK will clock the CPU onto an instruction load boundary.
	}

	HIL_TCLK(1);

	return (isInstrLoad()); // (As an internal check) Must exit on an instruction load boundary.
}


//-----------------------------------------------------------------------------
// Assumptions:
// 1) The processor is on an instruction load boundary. This implies that TCLK and R/W are high.
// 2) The program counter does not point into RAM.

STATUS_T writeWord(WORD Address, WORD Length, WORD* Buffer)
{
	int i;

	if (Length == 0)
	{
		return (STATUS_OK);
	}
	if (Address & 1) // Address must be even.
	{
		return (STATUS_ERROR);
	}
	if (isInstrLoad() != STATUS_OK) // (As an internal check) Must start on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}

	// Peripherals (especially the Watchdog timer as manipulated in restoreRegs()) must not be accessed using DATA QUICK
	// (since the Status Register can be restored prior to the use of DATA QUICK, and DATA QUICK will not work as
	// implemented as follows if the CPU is Off). Bottom line: Don't use DATA QUICK for peripherals.
	if (Address > LAST_PERIPHERAL_ADDR)
	{
		// Write memory (ram, information, program).

		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
		HIL_TCLK(0);
		//HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16((WORD) (Address - 2));
		HIL_TCLK(NEG_EDGE);
		HIL_TCLK(NEG_EDGE);
		HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
		if (HIL_JTAG_DR16(0) != Address - 2)
		{
			return (STATUS_ERROR);
		}
		// The next rising TCLK will clock the CPU onto an instruction load boundary.

		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(JMP_D); // Load "jmp $" instruction for use while HALT'ed.
		HIL_TCLK(NEG_EDGE);
		if (isInstrLoad() != STATUS_OK) // (As an internal check) Must be on an instruction load boundary.
		{
			return (STATUS_ERROR);
		}
		// The F12x2 requires that TCLK be low when HALT_JTAG is set.
		cntrlSigLowByte(CNTRL_SIG_HALT_JTAG); // Enable write to memory and halt the CPU (with the next TCLK).
		HIL_JTAG_IR_(IR_DATA_QUICK);
	
		for (i = 0; i < Length; i++)
		{
			HIL_JTAG_DR16(Buffer[i]);
			HIL_TCLK(NEG_EDGE);
		}
	}
	else
	{
		// Write peripherals. Don't use DATA_QUICK.
		HIL_JTAG_IR_(IR_DATA_16BIT);
		HIL_JTAG_DR16(JMP_D); // jmp $ for use while HALT'ed.
		HIL_TCLK(0);
		// The F12x2 requires that TCLK be low when HALT_JTAG is set.
		cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG); // Halt the CPU.
		HIL_TCLK(NEG_EDGE); // Stop CPU with this clock.
		cntrlSigLowByte(CNTRL_SIG_HALT_JTAG); // Enable write to peripherals.

		for (i = 0; i < Length; i++)
		{
			HIL_JTAG_IR_(IR_ADDR_16BIT);
			HIL_JTAG_DR16((WORD) (Address + i * 2));
			HIL_JTAG_IR_(IR_DATA_TO_ADDR);
			HIL_JTAG_DR16(Buffer[i]);
			HIL_TCLK(NEG_EDGE);
		}
	}

	cntrlSigLowByte(CNTRL_SIG_READ); // Release the CPU and disable memory write.
	// The next rising TCLK will clock the CPU onto an instruction load boundary.

	if (deviceHasJTAGBug())
	{
		WORD pc;

		HIL_JTAG_IR_(IR_ADDR_CAPTURE);
		pc = (WORD) HIL_JTAG_DR16(0);
		// Does the PC point into RAM? Note: Some devices do not completely decode the RAM address space.
		// Consequently, check if the PC is anywhere in the RAM space (and not just where the device has actual RAM).
		if ((pc >= deviceRamStart()) && (pc < deviceRamEnd()))
		{
			// Prevent RAM corruption by setting the PC to an address outside of RAM.
			HIL_JTAG_IR_(IR_DATA_16BIT);
			HIL_JTAG_DR16(MOV_IMM_PC); // Load the Program Counter.
			HIL_TCLK(NEG_EDGE);
			//HIL_JTAG_IR_(IR_DATA_16BIT);
			HIL_JTAG_DR16(ROM_ADDR);
			HIL_TCLK(NEG_EDGE);
			HIL_TCLK(NEG_EDGE);
			HIL_JTAG_IR_(IR_ADDR_CAPTURE); // Verify the Program Counter.
			if (HIL_JTAG_DR16(0) != ROM_ADDR)
			{
				return (STATUS_ERROR);
			}
		}
		// The next rising TCLK will clock the CPU onto an instruction load boundary.
	}

	HIL_TCLK(1);

//	return (isInstrLoad()); // (As an internal check) Must exit on an instruction load boundary.
    return(instrLoad()); // With the DTC, an extra cycle seems to be needed sometimes!?
}

//-----------------------------------------------------------------------------
// Assumptions:
// 1) The processor is on an instruction load boundary. This implies that TCLK and R/W are high.
// 2) The program counter does not point into RAM.

STATUS_T readByte(WORD Address, WORD Length, BYTE* Buffer)
{
	int i;

	if (Length == 0)
	{
		return (STATUS_OK);
	}
	if (isInstrLoad() != STATUS_OK) // (As an internal check) Must start on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}

	HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(JMP_D); // jmp $ for use while HALT'ed.
	HIL_TCLK(0);
	// The F12x2 requires that TCLK be low when HALT_JTAG is set.
	cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG); // Halt the CPU.
	HIL_TCLK(NEG_EDGE); // Stop CPU with this clock.
	cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG | CNTRL_SIG_BYTE); // Enable read of byte memory. 

	for (i = 0; i < Length; i++)
	{
		HIL_JTAG_IR_(IR_ADDR_16BIT);
		HIL_JTAG_DR16((WORD) (Address + i));
		HIL_JTAG_IR_(IR_DATA_TO_ADDR);
		HIL_TCLK(NEG_EDGE);
		Buffer[i] = (BYTE) (HIL_JTAG_DR16(0) & 0xff); // Ignore upper byte of value read.
	}

	cntrlSigLowByte(CNTRL_SIG_READ); // Release the CPU.
	// The next rising TCLK will clock the CPU onto an instruction load boundary.

	HIL_TCLK(1);

	return (isInstrLoad()); // (As an internal check) Must exit on an instruction load boundary.
}

//-----------------------------------------------------------------------------
// Assumptions:
// 1) The processor is on an instruction load boundary. This implies that TCLK and R/W are high.
// 2) The program counter does not point into RAM.

STATUS_T writeByte(WORD Address, WORD Length, BYTE* Buffer)
{
	int i;

	if (Length == 0)
	{
		return (STATUS_OK);
	}
	if (isInstrLoad() != STATUS_OK) // (As an internal check) Must start on an instruction load boundary.
	{
		return (STATUS_ERROR);
	}

	HIL_JTAG_IR_(IR_DATA_16BIT);
	HIL_JTAG_DR16(JMP_D); // jmp $ for use while HALT'ed.
	HIL_TCLK(0);
	// The F12x2 requires that TCLK be low when HALT_JTAG is set.
	cntrlSigLowByte(CNTRL_SIG_READ | CNTRL_SIG_HALT_JTAG); // Halt the CPU.
	HIL_TCLK(NEG_EDGE); // Stop CPU with this clock.
	cntrlSigLowByte(CNTRL_SIG_HALT_JTAG | CNTRL_SIG_BYTE); // Enable write to byte memory.

	for (i = 0; i < Length; i++)
	{
		HIL_JTAG_IR_(IR_ADDR_16BIT);
		HIL_JTAG_DR16((WORD) (Address + i));
		HIL_JTAG_IR_(IR_DATA_TO_ADDR);
		HIL_JTAG_DR8(Buffer[i]); // Upper byte of register will not be written.
		HIL_TCLK(NEG_EDGE);
	}

	cntrlSigLowByte(CNTRL_SIG_READ); // Release the CPU and disable memory write.
	// The next rising TCLK will clock the CPU onto an instruction load boundary.

	HIL_TCLK(1);

	return (isInstrLoad()); // (As an internal check) Must exit on an instruction load boundary.
}
