/**
 *   \file
 *   $Author: bpisupat $
 *   $Date: 2007/06/05 16:01:43 $
 *   $Revision: 1.1.1.1 $  
 *
 *   This code is derived from plan9port by Russ Cox which
 *   is in turn derived from the Plan 9 OS and covered by
 *   the Lucent Public License
 *   
*/


#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "thread.h"

#ifdef MSP430
#define assert(x) if (0) {}
#else
#include <assert.h>
#endif


#define otherop(op) (CHANSND+CHANRCV - (op))

/**
 *  Tests an alt structure for waiting matching action
 *  or available buffer space.
 *
 * \param a  The alt structure to check
 * \return   Returns if alt has matching action
 *
 */

/**
 * Computes rand()%%N
 * \param n The modulus N
 * \return  number in range 0..N-1
 */

static unsigned
randN(unsigned n)
{
static unsigned long next = 1;
next = (next * 1103515245 + 12345);
return (((next >> 16) * n) >> 16) & 65535;
}


static int
altcanexec(Alt *a)
{
  Alt *aq;
  Chanop op = otherop(a->op);
  Channel *c;
 
  if (a->op == CHANNOP)
    return 0;
  c = a->c;
  if (c->bufsize == 0)
    {
      for (aq = a->c->aq; aq; aq = aq->next)
	{
	  if (aq->op == op)
	    return 1;
	}
      return 0;
    }
  else
    {
      switch(a->op){
      case CHANSND:
	return c->nbuf < c->bufsize;
      case CHANRCV:
	return c->nbuf > 0;
      default:
	return 0;
      }
    }
}

static void
altqueue(Alt *a)
{
  Channel *c = a->c;
  a->next = c->aq;
  c->aq = a;
}

static void
altdequeue(Alt *a)
{
  Alt **aptr;
  Channel *c = a->c;
  
  for (aptr = &(c->aq); *aptr; aptr=&((*aptr)->next))
    {
      if (*aptr == a)
	{
	  *aptr = a->next;
	  return;
	}
    }
}

static void
altalldequeue(Alt *a)
{
  int i;
  for (i = 0; a[i].op != CHANEND && a[i].op!=CHANNOBLK; i++)
    if (a[i].op != CHANNOP)
      altdequeue(&a[i]);
}

static void
amove(void *dst, void *src, unsigned n)
{
  if (dst)
    {
      if (src == NULL)
	memset(dst, 0, n);
      else
	memmove(dst,src,n);
    }
}

static void 
altcopy(Alt *s, Alt *r)
{
  Channel *c;
  Alt *t;
  unsigned char *cp;

  if (s == NULL && r == NULL)
    return;
  assert(s != NULL);

  c = s->c;
  if (s->op == CHANRCV) 
    {
      t = s;
      s = r;
      r = t;
    }
  assert(s==NULL || s->op == CHANSND);
  assert(r==NULL || r->op == CHANRCV);

  /*
   * Channel is unbuffered or empty, copy directly
   */

  if (s && r && (c->nbuf == 0))
    {
      amove(r->v, s->v, c->elemsize);
      return;
    }

  /*
   * otherwise receive then send
   */

  if(r){
    cp = (unsigned char *) &c[1];
    cp += (c->off)*(c->elemsize);
    amove(r->v, cp, c->elemsize);
    --c->nbuf;
    if(++c->off >= c->bufsize)
      c->off = 0;
  }
  if(s){
    int p = c->off + c->nbuf;
    if (p >= c->bufsize)
      p -= c->bufsize;
    cp = (unsigned char *) &c[1];
    cp += (p * c->elemsize);
    amove(cp, s->v, c->elemsize);
    ++c->nbuf;
  }
}

static void
altexec(Alt *a)
{
  int i;
  Channel *c = a->c;
  Chanop op = otherop(a->op);
  Alt *an;
  Alt *other = NULL;
  int ready = 0;
 
  i = 0;
  for (an = c->aq; an; an = an->next)
    {
      if (an->op == op)
	{
	  ready++;
	  other = an;
	}
    }

  if (ready > 1)
    {
      int j = randN(ready);
      for (an = c->aq; an; an = an->next)
	{
	  if (an->op == op)
	    {
	      if (j-- == 0)
		{
		  other = an;
		  break;
		}
	    }
	}
    }
  if (other)
    {
      altcopy(a, other);
      other->thread->sel = other - other->thread->xalt;
      altalldequeue(other->thread->xalt);
      _threadready(other->thread);
      printf("thread %p out of block\n", other->thread);
    }
  else
    altcopy(a, NULL);
}

/**
 *  Schedule (and possibly perform) an alt operation.
 *
 * \param a  The alt array to schedule.  Must be terminated
 * by CHANEND or CHANNOBLK
 *
 * \return < 0 : the alt events are queued. currentThread state set to Rendezvous.
 * \return >-1 : the index of the executed event
 * \return Result is also placed in currentThread->sel
 */

int
chanalt(Alt *a)
{
  int i, chosen, ncan, n, canblock;
 
  Thread *t;
  int pri;

  chosen = 0;
  pri = splhi();
  
  for (i = 0, ncan = 0; (a[i].op != CHANEND) && (a[i].op != CHANNOBLK); i++)
    {
      if (altcanexec(&a[i]))
	{
	  ncan++;
	  chosen = i;
	}
    }

  n = i;  
  if (ncan > 1)
    {
      int j = randN(ncan);
      for (i = 0; i < n; i++)
	{
	  if (altcanexec(&a[i]))
	    if (j-- == 0)
	      {
		chosen = i;
		break;
	      }
	}
    }

  canblock = a[i].op == CHANEND;
  
  t = currentThread;
  t->xalt = a;
  
  if (ncan)
    {
      altexec(&a[chosen]);
      t->sel = &a[chosen] - a;
    }
  else if (canblock)
    {
      for(i = 0; i < n; i++) 
	{
	  if (a[i].op != CHANNOP)
	    {
	      a[i].thread = t;
	      altqueue(&a[i]);
	    }
	}
      t->state = Rendezvous;
      t->sel = -1;
    }
  else
    {
      t->sel = n;

    }
  splx(pri);
  return t->sel;
}


static int
_chanop(Channel *c, int op, void *p, int canblock)
{
	Alt *a = currentThread->a;

	a[0].c = c;
	a[0].op = op;
	a[0].v = p;
	a[1].op = canblock ? CHANEND : CHANNOBLK;
	return chanalt(a);
}

/**
 * Blocking send on a single channel
 *
 * \param c the channel to send on
 * \param v the data to send
 *
 * \return < 0 : event is queued. currentThread state set to Rendezvous.
 * \return   0 : send is completed
 * \return Result is also placed in currentThread->sel
 */


int
chansend(Channel *c, void *v)
{
	return _chanop(c, CHANSND, v, 1);
}

/**
 * Non-blocking send on a single channel
 *
 * \param c the channel to send on
 * \param v the data to send
 *
 * \return  0 : send is completed. 
 * \return  1 : send failed.
 * \return Result is also placed in currentThread->sel
 */

int
channbsend(Channel *c, void *v)
{
	return _chanop(c, CHANSND, v, 0);
}

/**
 * Blocking receive on a single channel
 *
 * \param c the channel to receive from
 * \param v the variable to receive data
 *
 * \return < 0 : event is queued. currentThread state set to Rendezvous. 
 * \return   0 : receive is completed.
 * \return Result is also placed in currentThread->sel
 */

int
chanrecv(Channel *c, void *v)
{
	return _chanop(c, CHANRCV, v, 1);
}

/**
 * Non-blocking receive on a single channel
 *
 *
 * \param c the channel to receive from
 * \param v the variable to receive data
 *
 * \return   0 : receive is completed
 * \return   1 : receive failed
 * \return Result is also placed in currentThread->sel
 */


int
channbrecv(Channel *c, void *v)
{
	return _chanop(c, CHANRCV, v, 0);
}
